/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     Logging
// @module  Cyrano Diagnostics | This module contains the implementation of
//          several diagnostics classes.  The particular classes in this module
//          are intended for viewing generated diagnostic information.  There
//          are classes for offline as well as online viewing.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Logging/CECDiagnostic.h>
#include <CyrHelper.inl>

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor for a trace object takes a string and parses it
//          into the relevant fields.
// @parm    The string to be parsed.
CECTrace::CECTrace( const char* trace )
{

    string str_parse(trace);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(SEPARATOR ,pipe_pos);
    m_time = CTTime( atol( str_parse.substr(pipe_pos,startpos-pipe_pos).c_str()));
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(SEPARATOR ,pipe_pos);
    m_node = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(SEPARATOR ,pipe_pos);
    m_user = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(SEPARATOR ,pipe_pos);
    m_executor = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(SEPARATOR ,pipe_pos);
    m_category = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(SEPARATOR ,pipe_pos);
    m_level = atol( str_parse.substr(pipe_pos,startpos-pipe_pos).c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(SEPARATOR ,pipe_pos);
    m_msg = str_parse.substr(pipe_pos,startpos-pipe_pos);
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor connects to a currently active logging object in
//          preparation for reading data from it.
// @parm    Pointer to CORBA helper functionality.
// @parm    The name of the log to connect to.
CECDiagnostic::CECDiagnostic( CORBAServer* pServer, const char* name )
{
    // Assign member variables
    m_server = pServer;

    // Connect to remote log object
    CORBA::Object_ptr obj =
        m_server->FindObjectWithName( CyrLog::CLASS_NAME, name );
    m_log = CyrLog::_narrow( obj );
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the first trace entry from the remote logging
//          object.  The object returned from this call must be deleted by the
//          calling code.
// @rdesc   Pointer to the first trace entry or NULL if one cannot be found.
CECTrace* CECDiagnostic::GetFirstTrace()
{
    CECTrace* trace = NULL;

    // Get first trace entry from log object
    CORBA::String_var entry = m_log->GetFirstTrace();
    if ( strlen( entry ) > 0 )
        trace = new CECTrace( entry );

    return trace;
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves the next trace entry from the remote logging
//          object.  The object returned from this call must be deleted by the
//          calling code.
// @rdesc   Pointer to the next trace entry or NULL if one cannot be found.
CECTrace* CECDiagnostic::GetNextTrace()
{
    CECTrace* trace = NULL;

    // Get next trace entry from log object
    CORBA::String_var entry = m_log->GetNextTrace();
    if ( strlen( entry ) > 0 )
        trace = new CECTrace( entry );

    return trace;
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor opens a log file and prepares to read data from it.
// @parm    The fully specified path name for the log file.
CECDiagnosticFile::CECDiagnosticFile( const char* filename )
{
    // Open specified filename
    m_filename = filename;
    m_repofile = false;
    m_file = new fstream( filename, std::ios::binary|std::ios::in );
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor reads a log file form the OpenSTARep.  The
//          file is first copied to a temporary location and then opend in the
//          same way as in the default constructor.
// @parm    Pointer to a OpenSTARep file.
CECDiagnosticFile::CECDiagnosticFile( CECOrbFile* pFile )
{
    // Transfer file contents to temporary file
    m_filename = CyrGetTempFileName();
    pFile->CheckOut( m_filename.c_str(), false );

    // Open temporary file
    m_repofile = true;
    m_file = new fstream( m_filename.c_str(), std::ios::binary|std::ios::in );
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor closes the open log file.  If the file had been
//          transferred from the OpenSTARep , then it is deleted.
CECDiagnosticFile::~CECDiagnosticFile()
{
    // Close file
    delete m_file;

    // If opened from OpenSTARep then delete temporary file
    if ( m_repofile )
        CyrDeleteFile( m_filename.c_str() );
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method resets the current file offset to the beginning of the
//          log file.  This effectively allows the iteration to begin again.
// @rdesc   Pointer to the first trace entry or NULL if one cannot be found.
CECTrace* CECDiagnosticFile::GetFirstTrace()
{
    // Reset current offset in file by re-opening file
    m_file->seekp(0L,std::ios::beg);

    return GetNextTrace();
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a single line of text from an open log file.
//          This line of text is passed into a new trace entry and parsed
//          automatically.
// @rdesc   Pointer to the next trace entry or NULL if one cannot be found.
CECTrace* CECDiagnosticFile::GetNextTrace()
{
    // Read a single line from the file
    string entry( "" );
    char c;
    do
    {
        if ( m_file->read( &c ,1) )
            entry += c;
        else
            return NULL;
    }
    while (( c != '\n' ) && ( !m_file->eof() ));

    // Create a new trace entry from this string
    return new CECTrace( entry.c_str() );
}

