/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     Logging
// @module  Cyrano Log | This module contains the implementation of the
//          Cyrano logging interface.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Logging/CECLog.h>
#include <CyrHelper.inl>
#include <CECSimpleParse.h>
///////////////////////////////////////////////////////////////////////////////
#include <fstream>
using std::fstream;
const char* LOG_SETTINGS_FILE = "TraceSettings";

////////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor creates an instance of the CORBA CyrLog
//          implementation.
// @parm    Pointer to CORBA helper functionality.
// @parm    Name of log object.
// @parm    Default level of all logging.
CECLog::CECLog( CORBAServer* pServer, const char* name, long level /*=0*/ )
{
    // Assign member variables
    m_server = pServer;
    m_name = name;

    // Create and register instance of log object
    m_log = new CyrLogFile( m_server, level );
	m_server->ActivateObject(m_log);
    m_server->NameObject( m_log->_this(), CyrLog::CLASS_NAME, name);
    m_delete = false;
}


////////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor is responsible for disposing of the CORBA log object
//          and also for deregistering it.
CECLog::~CECLog()
{
    // Get log filename
    string filename = m_log->GetFullPath();

    // Shutdown CORBA log object
    m_server->UnNameObject( CyrLog::CLASS_NAME, m_name.c_str() );
    m_log->_remove_ref();

    // Delete created file to prevent build up of unwanted files.
    if ( m_delete == true )
        CyrDeleteFile( filename.c_str() );
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method simply passes on a log request to the underlying CORBA
//          object.
// @parm    The category for this diagnostic message.
// @parm    The level of the diagnostic message.
// @parm    The diagnostic message.
void CECLog::AddTrace( const char* category, long level, const char* msg )
{
    m_log->AddTrace( category, level, msg );
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method stores the current contents of the log file into a
//          specified OpenSTARep folder.
// @parm    Pointer to the OpenSTARep folder.
void CECLog::Store( CECOrbFolder* pFolder )
{
    if ( pFolder != NULL )
    {
        // Create new log file object in OpenSTARep by copying the file across
        CECOrbFile* pFile;
        pFile = new CECOrbFile( pFolder, m_name.c_str(), "Diagnostic File" );


        // Extract filename and path from log object
        string path = m_log->GetFullPath();
        //pFile->Lock();

        pFile->CheckIn( path.c_str(), true );

        // Tidy memory
        delete pFile;

        // We now want to delete the log file
        m_delete = true;
    }
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method saves the current trace settings to a file in the root
//          folder of a OpenSTARep project.
// @parm    Pointer to a OpenSTARep project.
void CECLog::SaveTraceLevels( CECOrbProject* pProject )
{
    // Get root folder of project
    CECOrbFolder* pRoot = new CECOrbFolder( pProject );

    // Open temporary file for log details
    string filename = CyrGetTempFileName();
    fstream* pTempFile = new fstream( filename.c_str(), std::ios::binary|std::ios::out);
    if ( pTempFile->is_open() )
    {
        // Write log details to a single string
        TraceCat_var cats = m_log->GetTraceCategories();
        string entry( "" );
        char buf[20];
        for ( long i = 0; i < cats->length(); i++ )
        {
            // Build config string
            entry += cats[i];
            sprintf( buf, "|%d|", m_log->GetLevel( cats[i] ));
            entry += buf;
        }

        // Write string to file and close it
        pTempFile->write( entry.c_str(), entry.length() );
    }
    delete pTempFile;

    // Transfer file to OpenSTARep as the new log settings file.
    CECOrbFile* pFile;

    pFile = new CECOrbFile( pRoot, LOG_SETTINGS_FILE , "Trace Settings File" );
    pFile->CheckIn( filename.c_str(), true );

    // Tidy memory
    delete pFile;
    delete pRoot;

    // Delete trace level file
    CyrDeleteFile( filename.c_str() );
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method should be called once a connection to OpenSTARep is
//          established.  This connection will allow retrieval of current trace
//          settings.
// @parm    The project where all trace settings are stored.
void CECLog::Connect( CECOrbProject* pProject )
{
    // Retrieve current log levels
    if ( pProject != NULL )
        RestoreTraceLevels( pProject );
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This function retrieves the current trace settings from a OpenSTARep
//          project.
// @parm    Pointer to a OpenSTARep project.
void CECLog::RestoreTraceLevels( CECOrbProject* pProject )
{
    // Get root folder of project and determine if log settings file exists
    CECOrbFolder* pRoot = new CECOrbFolder( pProject );

    // Transfer file contents to temporary location
    string filename = CyrGetTempFileName();
    CECOrbFile* pFile = new CECOrbFile( pRoot, LOG_SETTINGS_FILE );
    pFile->CheckOut( filename.c_str(), false );

    // Open file and read contents
    fstream* pTempFile = new fstream( filename.c_str(), std::ios::in|std::ios::binary);
    string entry( "" );
    bool bEOF = false;
    char c;
    do
    {
        if ( pTempFile->read( &c ,1) )
            entry += c;
        else
            bEOF = true;
    }
    while ( !pTempFile->eof() && (!bEOF));


    // Add trace levels
    CECSimpleParse spse(entry);
    string cat = spse.getNext(SEPARATOR);
    // Extract fields
    while(!cat.empty())
    {
        long level = atol(spse.getNext(SEPARATOR).c_str());
        // Sret category level
        m_log->SetLevel(cat.c_str(),level);
        cat = spse.getNext(SEPARATOR);
    }



    // Tidy memory
    delete pTempFile;
    delete pFile;
    delete pRoot;

    // Delete temporary file
    CyrDeleteFile( filename.c_str() );
}
////////////////////////////////////////////////////////////////////////////////
// @mfunc   This function Set the logging ;level for a particular category
// @parm    category
// @parm    level
void CECLog::SetLevel( const char* category, long level)
{
    m_log->SetLevel(category,level);
}

