/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_BINASC                  */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains the routine for performing conversions      */
/*  from a Binary number string to ASCII.                    */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>         /* General PSL Macros */

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <string.h>

/*
 *  Function prototypes
 */
#include psl_common_file (psl_rtndef.h) /* Define ourselves */

/*
 *  Structure definitions
 */
#include psl_common_file (shrmsg.h) /* PSL messages */
#include psl_common_file (binasc_prmdef.h)
/* BINASC symbols */


/************************************************************************/
/*                                  */
/*  PSL_BINASC - Convert a Binary number string to ASCII        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*    iret = PSL_BINASC (number_ptr, length_ptr, radix_ptr,     */
/*               ascstr_ptr, flags_ptr);            */
/*                                  */
/*  Where:                              */
/*                                  */
/*    Inputs:                           */
/*      number  = Pointer to binary number to be converted.     */
/*      length  = Pointer to length of converted number string  */
/*      radix   = Pointer to conversion radix.          */
/*      flags   = Pointer to Format Action flags:           */
/*          Bit 0 set - Signed Conversion           */
/*          Bit 1 set - Leading Zeros (not Spaces)      */
/*          Bit 2 set - Left Justified          */
/*                                  */
/*    Outputs:                          */
/*      ascstr  = Pointer to converted string.          */
/*                                  */
/*    Returns:                          */
/*     iret     = Unsigned Long PSL Routine Return Status       */
/*                                  */
/*  This routine converts a Binary string into its ASCII        */
/*  equivalent.                         */
/*                                  */
/*  NOTE: The (binary) number input is treated as "unsigned"    */
/*        or "signed" depending on the conversion radix; hence  */
/*        its declaration as type VOID (see comment later).     */
/*                                  */
/************************************************************************/

unsigned long PSL_BINASC (void *number_ptr,
                          unsigned char *stringlen_ptr,
                          unsigned char *radix_ptr,
                          char *ascstr_ptr,
                          unsigned short *flags_ptr)

{
    /*
     *  Local Data.
     */
    unsigned long   iret;               /* General status return */
    unsigned long   chars_left;         /* Unconverted chars in string*/
    unsigned long   digit;              /* A converted digit to output*/
    unsigned long   move_cnt;           /* No. of chars to justify */
    unsigned long   unumber;            /* Unsigned variant of number */
    long   snumber;            /* Signed variant of number */

    static char trans_array [36] =      /* Radix tranlsation array */
        {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B',
         'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
         'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z'};

    char        *chr_ptr;           /* Pointer to current char */
    char        *dest_ptr;          /* Pointer to new destination */
    char        fill_chr;           /* Either " " (space) or "0" */

    /*
     *  Initialise return status.
     */
    iret = PSL_NORMAL;              /* Assume "success" */

    /*
     *  Initialise working pointers and counters.
     */
    move_cnt = 0;               /* NO chars to be justified */
    chars_left = *stringlen_ptr;        /* Unconverted chars in string */

    /*
     *  First, see if there is anywhere to put the result, and exit if not!
     */
    if (chars_left == 0)            /* NO room in output string? */
    {
        goto number_overflow;           /* Return error */
    }

    /*
     *  Check that the radix supplied is valid.
     */
    if ((*radix_ptr < 2) || (*radix_ptr > 36))  /* Invalid radix supplied? */
    {
        goto radix_error;           /* Yes: say so! */
    }

    /*
     *  Ensure, at this point that the first character of the output buffer is
     *  a space as we may jump straight to jusify (when converting zero) and will
     *  test to see if we have already inserted a '-'.
     */
    *ascstr_ptr = ' ';

    /*
     *  Now see if there is anything to convert!
     *
     *  If not, set the string to ZERO (right-justified), any justification
     *  requirements will be sorted out later.
     *
     * +--------------------------------------------------------------------+
     * |                    >>> Compatibility fudge <<<         |
     * |                                    |
     * |    Only radix 10 numbers that are flagged for SIGNED  conversion   |
     * |    are  treated  as "signed".  Any other numbers,  in any of the   |
     * |    other (valid) radixes are treated as "unsigned".        |
     * |                                    |
     * |    This  is for reasons of compatibility with PSL_ASCBIN,  who's   |
     * |    own compatibility with previous versions of that routine  has   |
     * |    had to be rationalised (see comment therein).           |
     * +--------------------------------------------------------------------+
     */
    chr_ptr = ascstr_ptr + chars_left - 1;  /* Point to last character */

    if ((*radix_ptr == 10)          /* If Radix is 10, */
            && ((*flags_ptr & BINASC_SIGN) != 0))  /* and a Signed Conversion, */
    {
        snumber = *((long *) number_ptr);   /* Cast number as "signed", */
        /* to allow us to see, */
        if (snumber == 0)           /* if it is ZERO,... */
        {
            *chr_ptr = '0';         /* Copy one ZERO there */
            move_cnt = 1;           /* Flag "1 char to justify" */
            chars_left -= 1;            /* Reduce count of chars left */
            goto justify;           /* Do any justification */
        }
    }
    else                    /* Its NOT considered "signed" */
    {
        unumber = *((unsigned long *) number_ptr);
        /* Cast number as "unsigned" */
        if (unumber == 0)           /* if it is ZERO,... */
        {
            *chr_ptr = '0';         /* Copy one ZERO there */
            move_cnt = 1;           /* Flag "1 char to justify" */
            chars_left -= 1;            /* Reduce count of chars left */
            goto justify;           /* Do any justification */
        }
    }

    /*
     *  Now see if we need a sign on the front.
     *
     *  NOTE that once any SIGNed Radix 10 numbers have been detected
     *  and any handling of the sign made ("negatives" are recorded in
     *  the output string), they are treated as positive and "directed"
     *  to the "unsigned" variable, making the ensuing code much simpler!
     */
    if ((*radix_ptr == 10)          /* If Radix is 10, */
            && ((*flags_ptr & BINASC_SIGN) != 0))  /* and a Signed Conversion, */
    {
        if (snumber < 0)            /* if number is Negative? */
        {
            *ascstr_ptr = '-';          /* Set first char to "Minus" */
            snumber *= -1;          /* Make it positive, for ease */
            chars_left -= 1;            /* Reduce count of chars left */
        }
        else                    /* Its positive! */
        {
            *ascstr_ptr = ' ';          /* Set first char to "Space" */
        }
        unumber = snumber;          /* Use the "unsigned" variable */
        /* to make code simpler */
    }
    else                    /* Its unsigned & not radix 10*/
    {
        *ascstr_ptr = ' ';          /* Set first char to "Space" */
    }

    /*
     *  For as long as the number has not been fully converted,  or, until
     *  a digit is encountered that cannot be converted -  convert another
     *  digit!
     *
     *  This is achieved by repeatedly (integer) dividing the input number 
     *  by the radix until zero is reached,  each time saving the quotient
     *  and translating the remainder to its character equivalent,  to  be
     *  added (backwards) to the output string.
     *
     *  NOTE: We have validated the radix by this time, so there is NO way
     *        that the "next digit" calculation (below) can yield a  value
     *        that is out-of-bounds regarding the translation array!
     */
    for (/* NO initialisation */;
                                unumber != 0;
                                chr_ptr--, chars_left--, move_cnt++)

    {
        if (chars_left == 0)            /* No space in output string? */
        {                   /* We have overflowed, so... */
            goto number_overflow;       /* Fill string with "*" */
        }

        digit = unumber % *radix_ptr;       /* Get the remainder... */
        /* i.e. next digit for string */
        unumber /= *radix_ptr;          /* (Integer) divide the Number*/
        /* by the Radix, for next time*/
        *chr_ptr = trans_array [digit];     /* Translate digit to char, */
        /* & copy to output string */

    }

    /*
     *  The number has been converted!
     *
     *  We must adjust the "current character" pointer as it will have been
     *  moved on one position by the preceding for loop (in its declaration's
     *  "update" section)!
     */
    chr_ptr += 1;                               /* Relocate current char */

    /*
     *  Now, see if we need to justify the field "left" or "right",
     *  and/or move the sign to the front of the number.
     *
     *  NOTE: 
     *      - The number string is already right-justified.
     *  - Only the "-" (negative) sign is output if "sign-ing", 
     *    and it will already have been placed in the output
     *    string, as the FIRST character.
     *  - Only relocate the sign character ("-") if Right-justifying.
     *  - Only check for "fill Leading-zeros" if Right-justifying.
     *  
     */
justify:
    fill_chr = ' ';             /* Default Fill char to " " */

    if ((*flags_ptr & BINASC_JUSTIFY) == 0) /* Should we Right justify? */
    {                       /* Yes... */
        if ((*flags_ptr & BINASC_FILL) != 0)    /* Should we lead Zero fill? */
        {                   /* No... */
            fill_chr = '0';         /* Set Fill character to "0"*/
        }
        else if (*ascstr_ptr == '-')        /* Impose a leading sign? */
        {                   /* Yes... */
            chr_ptr -= 1;           /* Point to preceeding char */
            *chr_ptr = '-';         /* Place it before the number */
        }
    }

    /*
     *  Now, justify or fill (with Blanks, or Zeroes), as necessary.
     */
    if ((*flags_ptr & BINASC_JUSTIFY) != 0) /* Move string to the left? */
    {
        dest_ptr = chr_ptr - chars_left;    /* Get destination for Move */
        (void) memmove (dest_ptr, chr_ptr, move_cnt);
        /* Do the move */

        dest_ptr += move_cnt;           /* Skip moved string for fill */
        (void) memset (dest_ptr, ' ', chars_left);
        /* Remove what was there */
    }
    else                    /* Fill with spaces or zeros */
    {
        dest_ptr = chr_ptr - chars_left;    /* Get destination for fill */
        (void) memset (dest_ptr, fill_chr, chars_left);
        /* Do the fill */
    }

    /*
     *  Exit here, returning status to the caller.
     */
end:
    return iret;

    /*
     *  Error handling.
     *  ==============
     *
     *  The Radix specified for the conversion is nonsensical (i.e. < 2, > 36).
     */
radix_error:
    iret = PSL_BADRADIX;            /* Set 'Bad Radix' error */
    goto return_asterisks;          /* Fill string with asterisks*/

    /*
     *  Number is too large to convert.
     */
number_overflow:
    iret = PSL_NUMTOOBIG;           /* Set 'Number Too Big' Error */
    goto return_asterisks;          /* Fill string with asterisks*/

    /*
     *  Tidy up following an error - return a string of asterisks!
     */
return_asterisks:
          (void) memset (ascstr_ptr, '*', *stringlen_ptr);
    /* Fill string with asterisks */
    goto end;                   /* and exit! */

    /*
     *  End of PSL_BINASC Routine.
     */
}
