/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 * scl_err_routines.c:
 * ===================
 *
 * This module contains TPA action routines used to report compilation errors.
 */

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines         */
#include "scl_rtndef.h"         /* SCL Routines         */
#include "scl_def_tbl_rtndef.h"     /* SCL definitions routines */
#include "scl_env_tbl_rtndef.h"     /* SCL environment routines */
#include "scl_cod_tbl_rtndef.h"     /* SCL code  routines       */
#include "scl_sub_tbl_rtndef.h"     /* SCL subexpression routines   */

static   void   STORE_ERROR (struct scl_com *scl_com_ptr,
                             unsigned long errnum);

static   void   UNDERLINE_ERROR (struct tpaarg *argblk_ptr);

/*
 *  Structure definitions.
 */
#include <shrmsg.h>         /* PSL Messages         */
#include <psldiag_prmdef.h>     /* Diagnostic Routine Symbols   */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block    */
#include <psl_tpatabdef.h>      /* TPARSE Table         */
#include <scpdef.h>         /* SCP Header record        */
#include "scl_tbldef.h"         /* SCL Table Definitions    */
#include <vtest_prmdef.h>       /* VTEST Parameters     */
#include "scl_limits.h"         /* SCL limits           */
#include "scl_comdef.h"         /* Common Data          */
#include "scl_tpacom.h"         /* Tparse common Data       */

/************************************************************************/
/*                                  */
/*    WARNING_CMND - Handle warning error               */
/*    =============                         */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  WARNING_CMND (argblk_ptr);                      */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Sets test-case warning flag.                        */
/*                                  */
/*    Returns:                              */
/*                                  */
/*    Description:                          */
/*                                  */
/*    This routine sets a flag to indicate that the line about to be    */
/*    may never get executed. This warning is output at the end of      */
/*    parsing the line.                             */
/*                                  */
/************************************************************************/
unsigned long   WARNING_CMND (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct scl_com *com_ptr;        /* Pointer to common data record*/
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* Dereference TPARSE pointer   */

    (*tpa_ptr).test_case_warnl = 1;
    return PSL_NORMAL;
}
/************************************************************************/
/*                                  */
/*    REPORT_WARNING - Handle warning error             */
/*    ==============                            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  REPORT_WARNING (argblk_ptr);                        */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Outputs a warning message.                              */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*    Description:                          */
/*  This routine outputs a warning message. It also maintains some  */
/*      information about lines never executed after a PASS/FAIL    */
/*  testcase command. (Not the nicest way of doing it I know!)  */
/*                                  */
/************************************************************************/

unsigned long   REPORT_WARNING (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct scl_com *com_ptr;        /* Pointer to common data record*/
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* Dereference TPARSE pointer   */

    /*
     *  If we have a command not executed then report the warning.
     */
    if ((*tpa_ptr).test_case_warnl == 1)
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_CMNDNOTEXEC);
        /* Set up diagnostic */
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).test_case_warnl = 0;
        (*tpa_ptr).test_fgl = 0;
    }

    /*
     * 
     *  If the warning is an unstarted sql block, report it.
     */
    if ((*tpa_ptr).sql_warnn)
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_NOSQLSTART);
        /* Set up diagnostic */
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).sql_warnn = PFALSE;
    }


    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    REPORT_ERROR - Handle reporting of an error           */
/*    =============                         */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  REPORT_ERROR (argblk_ptr, errnum);              */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Error stored in error array.                            */
/*  string count set to 0 if ignoring rest of line          */
/*                                  */
/*    Returns:                              */
/*                                  */
/*    Description:                          */
/*                                  */
/*    This routine is responsible for storing an Error in the array */
/*    of errors for the current source line and then underlining    */
/*    the current Token.                        */
/*                                  */
/************************************************************************/

void REPORT_ERROR (struct tpaarg *argblk_ptr,
                   unsigned long errnum)

{
    /*
     *  Local Data.
     */
    struct scl_com  *com_ptr;       /* Pointer to Common Data   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;

    /*
     *    Store the error in the array, if possible.
     */
    STORE_ERROR (com_ptr, errnum);

    /*
     *  Underline the Token in Error.
     */
    UNDERLINE_ERROR (argblk_ptr);

    /*
     *  Do we want to ignore the rest of the Line?
     */
    if ((*argblk_ptr).paraml == PSL_EXTTXT)
    {
        (*argblk_ptr).stringcntl = 0;
    }

    return;
}

/************************************************************************/
/*                                  */
/*    STORE_ERROR - Store an Error                  */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  STORE_ERROR (com_ptr, errnum);                  */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Error stored in error array.                    */
/*  Error count incremented                     */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                      */
/*    This routine is responsible for storing an Error in the array */
/*    of errors for the current source line and incrementing the    */
/*    appropriate counters.                     */
/*                                  */
/************************************************************************/

static void STORE_ERROR (struct scl_com *com_ptr,
                         unsigned long errnum)

{
    /*
     *  Local Data.
     */
    unsigned long   jsub;           /* Loop Control         */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    tpa_ptr = (*com_ptr).tpa_ptrps; /* Dereference TPARSE pointer   */

    /*
     *    Flag the fact that this command has an error in it. 
     *
     *    THIS WILL TURN OFF ALL FUTURE COMMAND ENCODING
     */
    (*tpa_ptr).error_in_commandn = PTRUE;

    /*
     *    Check to see if we have already filled the error array.
     */
    if  ((*tpa_ptr).errors_in_linel != MAX_ERRORS_IN_LINE)
    {                   /* If still space for errors    */

        /*
         *    Loop through the error array to make sure that the new error
         *    is not already in there. If it is, we just underline it.
         */
        for (jsub = 0; jsub < (*tpa_ptr).errors_in_linel; jsub++)
        {
            if ((*tpa_ptr).error_arrayl5 [jsub] == errnum) goto end;
        }

        /*
         *    If we get here the error is not already in the array so we insert it
         *    in the first FREE slot and count it.
         */
        (*tpa_ptr).error_arrayl5 [(*tpa_ptr).errors_in_linel] = errnum;

        (*tpa_ptr).errors_in_linel += 1;
    }

    /*
     *    Increment the error count before returning.
     */
end:
    (*tpa_ptr).error_countl += 1;

    return;
}

/************************************************************************/
/*                                  */
/*    UNDERLINE_ERROR - Underline the current Token         */
/*    ===============                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  UNDERLINE_ERROR (argblk_ptr);                   */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Current token underlined in the error line          */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                      */
/*    This routine is responsible for underlining the current Token.    */
/*    The ERROR_LINE is set up with asterisks in the correct places.    */
/*                                  */
/************************************************************************/

static void UNDERLINE_ERROR (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long    offset;        /* Character String Offset  */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Dereference COMMON pointer   */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* Dereference TPARSE pointer   */

    /*
     *    Calculate the offset of the character pointed to by token_ptr in
     *    the parsable_line buffer.
     */
    offset = (unsigned long) (*argblk_ptr).token_ptrpc -
             (unsigned long) (*tpa_ptr).parsable_linec136;

    /*
     *    Set the corresponding charcaters in the error_line buffer to
     *    asterisks.
     */
    (void) memset ((*tpa_ptr).error_linec136 + offset
                   ,'*'
                   ,(*argblk_ptr).tokencntl);

    /*
     *    Set the flag that indicates that an Underline Line is required.
     */
    (*tpa_ptr).underline_flagl = 1;

    return;
}

/************************************************************************/
/*                                  */
/*    ERROR_BLANK - Report error and skip to next blank     */
/*    ============                          */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = ERROR_BLANK (argblk_ptr);                */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Token extended to next non-printing character or end-of-line    */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                      */
/*    This routine expands the current token to the next non-printing   */
/*    character, space or 'End of Line', then underlines it and reports */
/*    the error, whose message number is passed in PSLTPA$L_PARAM(AP).  */
/*                                  */
/************************************************************************/

unsigned long   ERROR_BLANK (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long    tokencnt;      /* Saved token counter      */
    char    *wrk_ptr;      /* Work pointer to token    */

    tokencnt  = (*argblk_ptr).tokencntl; /* Copy token count        */
    wrk_ptr = (*argblk_ptr).token_ptrpc + tokencnt;
    /* Workl pointer to token pointer */

    /*
     *    Check the next character in the token to see if it is one of
     *  the designated terminators, i.e. a non-printing character,
     *  space or 'End of Line'.
     */
    while (((*argblk_ptr).stringcntl > 0) /* While not at end of string */
            &&     (isprint (*wrk_ptr) != 0))   /* ...and printing character    */
    {
        tokencnt += 1;          /* Update local token count */
        wrk_ptr += 1;           /* ...and local pointer     */
        (*argblk_ptr).string_ptrpc += 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl -= 1;  /* ...and TPARSE string count   */
    }

    /*
     *    Token extension complete. Update token count, underline the
     *    current token and report error.
     */
    (*argblk_ptr).tokencntl = tokencnt; /* Point TPARSE to current count */
    REPORT_ERROR (argblk_ptr, (*argblk_ptr).paraml);
    /* Pass to underline routine    */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    ERROR_CHR_CMND - Report error and skip past a following '>'   */
/*    ===============                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = ERROR_CHR_CMND (argblk_ptr);             */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Token extended to the next '>' character or end-of-line     */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                      */
/*    This routine is used to report an error having parsed the first   */
/*    two characters of a character command (i.e. ~<) successfully.     */
/*                                  */
/*    It expands the current token to a '>' character or the end of the */
/*    source line, and then it underlines the token & reports the error */
/*    passed in PSLTPA$L_PARAM(AP).                 */
/*                                  */
/*    If the terminator is 'end of line' AND the last character is the  */
/*    continuation character '-', it is NOT included in the token so    */
/*    it can be picked up in the TPARSE state tables.           */
/*                                  */
/************************************************************************/

unsigned long   ERROR_CHR_CMND (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long    tokencnt;      /* Work token counter       */
    char    *wrk_ptr;      /* Work pointer to token    */

    /*
     *    Save the current token count and set a pointer to point to the next
     *    character in the token being expanded.
     */
    tokencnt  = (*argblk_ptr).tokencntl; /* Copy token count        */
    wrk_ptr = (*argblk_ptr).token_ptrpc + tokencnt;
    /* Work pointer to token pointer */

    /*
     *    Check the next character in the token to see if it is a '>' character
     *    or 'End of Line'.
     */
    do
    {
        tokencnt += 1;          /* Update local token count */
        wrk_ptr += 1;           /* ...and local pointer     */
        (*argblk_ptr).string_ptrpc += 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl -= 1;  /* ...and TPARSE string count   */
    }
    while (((*argblk_ptr).stringcntl > 0) /* While not at end of string */
            &&     (*(wrk_ptr-1) != '>'));  /* ...and last token not '>'    */

    /*
     *    Token extension complete. Check to see if we completed because the last
     *    cheracter was either the delimiter character to a quoted string or was
     *    a continuation character.
     */
    if (((*argblk_ptr).stringcntl == 0) /* If we are at the end of string */
            &&  ((*(wrk_ptr-1) == '-') || (*(wrk_ptr-1) == '&')))
        /* ...and we have a continuation */
    {
        tokencnt -= 1;          /* Decrement local token count  */
        (*argblk_ptr).string_ptrpc -= 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl += 1;  /* ...and inc TPARSE string count */
    }

    /*
     *    Update the tioken count in the argument block and report the error
     */
    (*argblk_ptr).tokencntl = tokencnt; /* Point TPARSE to current count */
    REPORT_ERROR (argblk_ptr, (*argblk_ptr).paraml);
    /* Pass to underline routine    */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    ERROR_CMND - Report error and skip to designated terminator   */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = ERROR_CMND (argblk_ptr);                 */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Token extended to end-of-line                   */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                      */
/*    This routine is called when an error is detected in a command */
/*    and the rest of the command is to be ignored. The routine         */
/*    simply calls ERROR_LINE to expand the current token to include    */
/*    all characters remaining on the line, underline the       */
/*    characters in the token and report an error.          */
/*                                  */
/************************************************************************/
unsigned long ERROR_CMND (struct tpaarg *argblk_ptr)
{
    (*argblk_ptr).tokencntl =
        (*argblk_ptr).tokencntl + (*argblk_ptr).stringcntl;
    /* Expand token to end of line  */
    (*argblk_ptr).string_ptrpc =
        (*argblk_ptr).tokencntl + (*argblk_ptr).token_ptrpc;
    /* ...and expand string pointer */
    (*argblk_ptr).stringcntl = 0;

    REPORT_ERROR (argblk_ptr, (*argblk_ptr).paraml);
    /* Pass to underline routine    */
    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    ERROR_LINE - Report Error and Ignore Rest of Line         */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                           */
/*                                  */
/*  iret = ERROR_LINE (argblk_ptr)                  */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Token pointers and count expanded to end of the line        */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                  */
/*    This routine is called when an error is detected and the      */
/*    rest of the line is to be ignored. The routine first expands  */
/*    the current token to include all characters remaining on the  */
/*    line. It then underlines the characters in the token, reports */
/*    an error and reads the next source line.              */
/*                                  */
/************************************************************************/

unsigned long   ERROR_LINE (struct tpaarg *argblk_ptr)
{
    (*argblk_ptr).tokencntl =
        (*argblk_ptr).tokencntl + (*argblk_ptr).stringcntl;
    /* Expand token to end of line  */
    (*argblk_ptr).string_ptrpc =
        (*argblk_ptr).tokencntl + (*argblk_ptr).token_ptrpc;
    /* ...and expand string pointer */
    (*argblk_ptr).stringcntl = 0;

    REPORT_ERROR (argblk_ptr, (*argblk_ptr).paraml);
    /* Pass to underline routine    */

    /*
     *    We have terminated the line. Read in a new one.
     */
    (void) SCL_GET_LINE (argblk_ptr);   /* Read in a new line       */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    ERROR_MESSAGE - Report error message only.            */
/*    =============                             */
/*                                  */
/*                                  */
/*    Calling Convention:                           */
/*                                  */
/*  iret = ERROR_MESSAGE (argblk_ptr)               */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Error stored in the error array.                            */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                  */
/*    This routine reports the error passed in (*argblk_ptr).paraml     */
/*    It is used whenever a message needs to be reported, but no    */
/*    characters in the source line underlined.             */
/*                                  */
/*    This routine does not actually perform any function other than    */
/*    saving the error message number into the array of errors and  */
/*    incrementing the count of errors encountered.         */
/*                                  */
/************************************************************************/

unsigned long   ERROR_MESSAGE (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct scl_com *com_ptr;        /* Pointer to common data record*/

    /*
     *    De-reference the pointer to the common data record and call a routine
     *    to store the error whose value is passed as the TPASRE parameter
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Dereference COMMON pointer   */
    STORE_ERROR (com_ptr, (*argblk_ptr).paraml);
    /* Store error no underline */
    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    ERROR_OPT - Report Error in Option                */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                           */
/*                                  */
/*  iret = ERROR_OPT (argblk_ptr)                   */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Token expanded to the first ',' or ')' or Tab or to end of line */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                  */
/*    This routine is called when an error is detected when parsing a   */
/*    command or command option. The routine first expands the current  */
/*    token to include all characters up to the next comma      */
/*    space, or Tab character or to the end of the line.        */
/*                                  */
/*    If the terminator is 'End of Line' AND the last character is  */
/*    the continuation character '-', it is NOT included in the token   */
/*    so that it can be picked up in the TPARSE state tables.       */
/*                                  */
/************************************************************************/

unsigned long   ERROR_OPT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long   tokencnt;       /* Work token pointer       */
    char   *wrk_ptr;       /* Work pointer to token    */

    /*
     *    Save the current token count and set a pointer to the next character
     *    in the current token.
     */
    tokencnt  = (*argblk_ptr).tokencntl; /* Copy token count        */
    wrk_ptr = (*argblk_ptr).token_ptrpc + tokencnt;
    /* Workl pointer to token pointer */

    /*
     *    Check the next character in the token to see if it is one of
     *    the designated terminators, i.e. a comma, Tab, Space or end-of-line
     */
    while (((*argblk_ptr).stringcntl > 0) /* While not at end of string */
            &&     (*wrk_ptr != ',')        /* ...and current token not ',' */
            &&     (*wrk_ptr != ')')        /* ...and current token not ' ' */
            &&     (*wrk_ptr != '\t'))      /* ...and current token not TAB */
    {
        tokencnt += 1;          /* Update local token count */
        wrk_ptr += 1;           /* ...and local pointer     */
        (*argblk_ptr).string_ptrpc += 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl -= 1;  /* ...and TPARSE string count   */
    }

    /*
     *    Token extension complete. Update token count and report error.
     */
    (*argblk_ptr).tokencntl = tokencnt; /* Point TPARSE to current count */
    REPORT_ERROR (argblk_ptr, (*argblk_ptr).paraml);
    /* Pass to underline routine    */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    ERROR_PREMEOF - Check for unexpected End of File          */
/*    ==============                            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = ERROR_PREMEOF (argblk_ptr);              */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Premature end of file reported if end-of-file set.      */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL - If end-of-file                 */
/*  0 - If not end-of-file.                     */
/*                                  */
/*    Description:                          */
/*                                  */
/*    This routine checks the End of File flag and, if set, reports */
/*    an error.                             */
/*                                  */
/************************************************************************/

unsigned long   ERROR_PREMEOF (struct tpaarg *argblk_ptr)

{
    /*
     *  Local Data.
     */
    struct   scl_com *com_ptr;      /* Pointer to Common Data   */
    struct   scl_tpa *tpa_ptr;      /* Pointer to Common Data   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Get pointer to common data   */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* ...and then to TPARSE data   */

    /*
     *    If the EOF flag is set then report a premature end-of-file message.
     */
    if (psl_test_status ((*com_ptr).eof_flagl))
        /* End of File flag set?    */
    {
        (*tpa_ptr).error_countl += 1;   /* Increment Error Count    */
        PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                     1,
                     PSL_PREMEOF);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        return PSL_NORMAL;      /* ...but return success    */
    }
    else
    {
        return 0;
    }
}

/************************************************************************/
/*                                  */
/*    ERROR_PRM - Report Error in Parameter             */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                           */
/*                                  */
/*  iret = ERROR_PRM (argblk_ptr)                   */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Token expanded to next comma, closing bracket or end-of-line    */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                  */
/*    This routine is called when an error is detected when parsing a   */
/*    parameter in a parameter list. The routine first expands the  */
/*    current token to include all characters up to the next comma, */
/*    right bracket ')' or 'End of Line'. It then underlines all    */
/*    non-blank characters in the token and reports an error.       */
/*                                  */
/*    If the terminator is 'End of Line' AND the last character is  */
/*    the continuation character '-', it is NOT included in the token   */
/*    so that it can be picked up in the TPARSE state tables.       */
/*                                  */
/************************************************************************/

unsigned long   ERROR_PRM (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long   tokencnt;       /* Work token count     */
    char   *wrk_ptr;       /* Work pointer to token    */

    /*
     *    Save the current token count and set a pointer to the next character in
     *    the current token.
     */
    tokencnt  = (*argblk_ptr).tokencntl; /* Copy token count        */
    wrk_ptr = (*argblk_ptr).token_ptrpc + tokencnt;
    /* Work pointer to token pointer */

    /*
     *    Check the next character in the token to see if it is one of
     *    the designated terminators, i.e. a comma, right bracket or 'End of Line'.
     */
    while (((*argblk_ptr).stringcntl > 0) /* While not at end of string */
            &&     (*wrk_ptr != ',')        /* ...and current token not '>' */
            &&     (*wrk_ptr != ')'))       /* ...and current token not '>' */
    {

        /*
         *    None of the terminators - consume the token.
         */
        tokencnt += 1;          /* Update local token count */
        wrk_ptr += 1;           /* ...and local pointer     */
        (*argblk_ptr).string_ptrpc += 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl -= 1;  /* ...and TPARSE string count   */
    }

    /*
     *    Token extension complete. Check to see if we completed because the last
     *    cheracter was either the delimiter character to a quoted string or was
     *    a continuation character.
     */
    if (((*argblk_ptr).stringcntl == 0) /* If we are at the end of string */
            &&  ((*(wrk_ptr-1) == '-') || (*(wrk_ptr-1) == '&')))
        /* ...and we have a continuation */
    {
        tokencnt -= 1;          /* Decrement local token count  */
        (*argblk_ptr).string_ptrpc -= 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl += 1;  /* ...and inc TPARSE string count */
    }

    /*
     *    Update the token count in the argument block and report the error
     */
    (*argblk_ptr).tokencntl = tokencnt; /* Point TPARSE to current count */
    REPORT_ERROR (argblk_ptr, (*argblk_ptr).paraml);
    /* Pass to underline routine    */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    ERROR_TOKEN - Report Error for current token          */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                           */
/*                                  */
/*  iret = ERROR_TOKEN (argblk_ptr)             */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Error reported.                                             */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                  */
/*    This routine underlines the current token and reports the error,  */
/*    whose message number is passed in PSLTPA$L_PARAM(AP).     */
/*                                  */
/************************************************************************/

unsigned long   ERROR_TOKEN (struct tpaarg *argblk_ptr)

{
    REPORT_ERROR (argblk_ptr, (*argblk_ptr).paraml);
    /* Pass to underline routine    */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    ERROR_VALUE - Report Error in Value               */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                           */
/*                                  */
/*  iret = ERROR_VALUE (argblk_ptr)             */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Token extended to the next comma or end-of-line         */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                      */
/*    This routine is used to report an error in a value in a value */
/*    list. It first expands the current token to next value 'delimter' */
/*    i.e. ',', '-', ')' or end of line. It then underlines token and   */
/*    reports the error passed in PSLTPA$L_PARAM(AP).           */
/*                                  */
/*    If the terminator is 'end of line' AND the last character is the  */
/*    continuation character '-' it is NOT included in the token so     */
/*    it can be picked up in the TPARSE state tables.           */
/*                                  */
/************************************************************************/

unsigned long   ERROR_VALUE (struct tpaarg *argblk_ptr)

{
    /*
     *  Local Data.
     */
    unsigned long   tokencnt;       /* Work token count     */
    char   *wrk_ptr;       /* Work pointer to token    */

    /*
     *    Save the current token count and set a pointer to the next character in
     *    the current token.
     */
    tokencnt  = (*argblk_ptr).tokencntl; /* Copy token count        */
    wrk_ptr = (*argblk_ptr).token_ptrpc + tokencnt;
    /* Work pointer to token pointer */

    /*
     *    Check the next character in the token to see if it is one of
     *    the designated terminators, i.e. a comma, right bracket or 'End of Line'.
     
     */
    while (((*argblk_ptr).stringcntl > 0) /* While not at end of string */
            &&     (*wrk_ptr != ','))       /* ...and current token not ',' */
    {

        /*
         *    None of the terminators - consume the token.
         */
        tokencnt += 1;          /* Update local token count */
        wrk_ptr += 1;           /* ...and local pointer     */
        (*argblk_ptr).string_ptrpc += 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl -= 1;  /* ...and TPARSE string count   */
    }

    /*
     *    Token extension complete. Check to see if we completed because the last
     *    cheracter was either the delimiter character to a quoted string or was
     *    a continuation character.
     */
    if (((*argblk_ptr).stringcntl == 0) /* If we are at the end of string */
            &&  ((*(wrk_ptr-1) == '-') || (*(wrk_ptr-1) == '&')))
        /* ...and we have a continuation */
    {
        tokencnt -= 1;          /* Decrement local token count  */
        (*argblk_ptr).string_ptrpc -= 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl += 1;  /* ...and inc TPARSE string count */
    }

    /*
     *    Update the token count in the argument block and report the error
     */
    (*argblk_ptr).tokencntl = tokencnt; /* Point TPARSE to current count */
    REPORT_ERROR (argblk_ptr, (*argblk_ptr).paraml);
    /* Pass to underline routine    */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    ERROR_LIST - Report Error in Value List/Range         */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                           */
/*                                  */
/*  iret = ERROR_LIST (argblk_ptr)                  */
/*                                  */
/*    Input arguments:                          */
/*  argblk_ptr - Pointer to the TPARSE argument block       */
/*                                  */
/*    Outputs:                              */
/*  Token extended to the next ")" or end-of-line           */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL                          */
/*                                  */
/*    Description:                          */
/*                                      */
/*    This routine is used to report an error in a value list       */
/*    It first expands the current token to the value list delimiter    */
/*    i.e. ')' or end of line. It then underlines the token and     */
/*    reports the error passed in PSLTPA$L_PARAM(AP).           */
/*                                  */
/*    If the terminator is 'end of line' AND the last character is the  */
/*    continuation character '-' it is NOT included in the token so     */
/*    it can be picked up in the TPARSE state tables.           */
/*                                  */
/************************************************************************/

unsigned long   ERROR_LIST (struct tpaarg *argblk_ptr)

{
    /*
     *  Local Data.
     */
    unsigned long   tokencnt;       /* Work token count     */
    char   *wrk_ptr;       /* Work pointer to token    */

    /*
     *    Save the current token count and set a pointer to the next character in
     *    the current token.
     */
    tokencnt  = (*argblk_ptr).tokencntl; /* Copy token count        */
    wrk_ptr = (*argblk_ptr).token_ptrpc + tokencnt;
    /* Work pointer to token pointer */

    /*
     *    Check the next character in the token to see if it is one of
     *    the designated terminators, i.e. a ')' or 'End of Line'.
     */
    while (((*argblk_ptr).stringcntl > 0) /* While not at end of string */
            &&     (*(wrk_ptr-1) != ')'))   /* ...and current token not ')' */
    {

        /*
         *    None of the terminators - consume the token.
         */
        tokencnt += 1;          /* Update local token count */
        wrk_ptr += 1;           /* ...and local pointer     */
        (*argblk_ptr).string_ptrpc += 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl -= 1;  /* ...and TPARSE string count   */
    }

    /*
     *    Token extension complete. Check to see if we completed because the last
     *    character was either the delimiter character to a quoted string or was
     *    a continuation character.
     */
    if (((*argblk_ptr).stringcntl == 0) /* If we are at the end of string */
            &&  ((*(wrk_ptr-1) == '-') || (*(wrk_ptr-1) == '&')))
        /* ...and we have a continuation */
    {
        tokencnt -= 1;          /* Decrement local token count  */
        (*argblk_ptr).string_ptrpc -= 1;/* ...and TPARSE string pointer */
        (*argblk_ptr).stringcntl += 1;  /* ...and inc TPARSE string count */
    }

    /*
     *    Update the token count in the argument block and report the error
     */
    (*argblk_ptr).tokencntl = tokencnt; /* Point TPARSE to current count */
    REPORT_ERROR (argblk_ptr, (*argblk_ptr).paraml);
    /* Pass to underline routine    */
    return PSL_NORMAL;          /* ...and return success    */
}
