/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// AtvfRecRule.cpp
/**
 * Define methods for the AtvfRecRule class
 **/

// Local includes
#include "AtvfRecRule.h"
#include "mdbexcept.h"

/**
 * Constructor 
 **/
AtvfRecRule::AtvfRecRule(const string& token,
                         bool hasValue,
                         bool leadWSpace)
        :   _key(token),
        _hasValue(hasValue),
        _leadWSpace(leadWSpace)
{
    _token = token;
    // validate the rule, throw exception if not valid.
    // should at least one non-whitespace character.
    string::size_type firstNonWspace = _token.find_first_not_of(" \t");
    if(firstNonWspace == string::npos)
    {
        throw mdbParseException("No non-whitespace in token");
    }
    // and no whitespace after first non-whitespace
    if(_token.find_first_of(" \t", firstNonWspace) != string::npos)
    {
        throw mdbParseException("Whitespace within token");
    }
}

/**
 * Access to default token for this rule
 **/
string AtvfRecRule::token() const
{
    return _token;
}

/**
 * Access to this rules key
 **/
AtvfKey AtvfRecRule::key() const
{
    return _key;
}

/**
 * Access to fact if this Record type has a value
 **/
bool AtvfRecRule::hasValue() const
{
    return _hasValue;
}

/**
 * Access to fact of if Value associated with record can have
 * leading whitespace
 **/
bool AtvfRecRule::leadWSpace() const
{
    return _leadWSpace;
}

/**
 * Perform the default action of parsing input any descendents
 * which want different behaviour should overide.
 * Default behaviour is to return from current position in string
 * to the end of the line as the value.
 **/
string* AtvfRecRule::parseInput(AtvfInput& in) const
{
    return in.getLine();
}

/**
 * Save the given value in the format of this rule
 **/
void AtvfRecRule::saveValue(const string& value, ostream& out) const
{
    if(!_hasValue) return;
    if(value.size() > 0)
    {
        string::size_type pos = value.find_first_of(" \t");
        if(pos != 0) out << " ";
        out << value;
    }
}

/**
 * Define operator << to allow AtvfRecRule to be output.
 **/
ostream& operator<<(ostream& os, const AtvfRecRule& r)
{
    os << r.token() << "(" << r.type() << ")";
    return os;
}
