/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// AtvfRule.cpp
/**
 * Define the Methods for the AtvfRuleMap class
 * This class is a container for individual AtvfRecordRules that
 * make up a specific type of atvf file and any other info to define
 * the file type as a whole.  The version and records are allocated
 * and returned to the caller whose responsibility it is to delete them
 * when finished with.
 *
 * No validation of 'pointer to input stream' is done here in the
 * of this class that receive this.
 **/


#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
// Local includes
#include "AtvfRule.h"
#include "mdbexcept.h"

/**
 * Constructor - with string that identifies the atvf file type
 **/
AtvfRule::AtvfRule(const string& type)
        :   _type(type)
{}

/**
 *  Clear method to remove the rules content
 **/
void AtvfRule::clearRule()
{
    _rules.erase(_rules.begin(), _rules.end());
}


/**
 * Mutator to add a Record rule to the definition.
 * Record rules should be checked for uniqueness - it is OK to
 * redefine a rule associated with a key with the same rule (No
 * action) but not OK to change the rule associated with a key.
 **/
void AtvfRule::add(const SharedPtr<AtvfRecRule>& ruleP)
{
    // Check to see if we already have this rule
    rule_map_t::const_iterator p = _rules.find(ruleP->key());
    if(p != _rules.end())
    {
        // Already exists
        if((ruleP->type()).compare(p->second->type()) != 0)
        {
            // problem if it is a different type to existing rule
            throw mdbParseException("Incompatible rule definition");
        }
    }
    else
    {
        // Add the rule to the map and exit
        _rules.insert(rule_map_t::value_type(ruleP->key(), ruleP));
    }

    return;
}

/**
 * Accessor to validate input stream contains the type string at its
 * current postion and extract and return what follows on the line.
 **/
string* AtvfRule::extractHead(AtvfInput& in) const
{
    if(!in.getMatch(_type)) return NULL;

    // rest of line is what we return.
    return in.getLine();
}

/**
 * Accessor to output, to the given stream, a header record
 * for this atvf type.
 **/
void AtvfRule::saveHead(const string& version, ostream& out) const
{
    out << _type << version << endl;
}
/**
 * Accessor to extract a type of Atvf Record from the input stream at
 * the current positon.  Use our rules to decide what type of Atvf
 * Record we create.
 * Return NULL if we reach end of the input.
 **/
AtvfRecord* AtvfRule::extractRecord(AtvfInput& in) const
{
    string* tokP;
    tokP = in.getToken();
    if(tokP == NULL) return NULL; // signifies end of input.

    // Try to find the rule associated with the token.
    rule_map_t::const_iterator p = _rules.find(AtvfKey(*tokP));

    // Can't find a rule create and return Atvf Record with input line.
    if(p == _rules.end())
    {
        if(in.moreOnLine())
        { // Line not finished
            string* rofP = in.getLine();
            *tokP += *rofP;
            delete rofP;
        }
        return new AtvfRecord(tokP);
    }

    // Create an Atvf Record with pointer to rule,
    AtvfRecord* recP = new AtvfRecord(p->second, tokP);
    if(in.moreOnLine())
    {
        recP->getValue(in);
    }

    return recP;
}


ostream& operator<<(ostream& os, const AtvfRule& r)
{
    for (rule_map_t::const_iterator _ruleIt =  r._rules.begin();
            _ruleIt != r._rules.end(); ++_ruleIt)
        (_ruleIt ==  r._rules.begin()) ?
        os << *(_ruleIt->second): os << endl << *(_ruleIt->second);

    return os;
}



