/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TScriptRule.h
#ifndef _CYR_TSCRIPTRULE_H
#define _CYR_TSCRIPTRULE_H
/**
 * Class Definititon for transaction script rule class.
 * This class holds all the rules that describe the form and function
 * of a transaction script (TScript), These rules are created by parsing
 * an 'odl' (Object defintion language) file which is a form of atvf, which
 * is passed in to the constructor of this class.  The rules that are created
 * from this file describe a Transaction script and its file (serialized) 
 * format which is again based on the atvf style. 
 **/

#include <string>
#include <vector>
#ifdef unix
#elif defined (__VMS)
#elif defined (WIN32)
#pragma warning(disable:4786)
using namespace std;
class TScriptRule;
ostream& operator << (ostream& os, TScriptRule& tS);
#else //unknown
#error Unknown Platform
#endif // unknown

// local includes
#include "AtvfRule.h"
#include "TScriptHead.h"
#include "TRecord.h"
#include "TRecCreator.h"
#include "CConfigTable.h"
// Required for inclusion of the ODL atvf rule in class
#include "AtvfRealRecRules.h"

typedef vector<AtvfKey> keyvect_t;
typedef vector<string> strvect_t;
//  To limit parameter scope to a single file multiple AtvfRule's are reqired, one
//  per #include odl, a 'record token' to rule lookup, and a separate rule for
//  'record tokens' so that no rule decision is required to extract a 'record token'.
typedef map<AtvfKey, AtvfRule*, less<AtvfKey> >  toknrulemap_t;
typedef vector<AtvfRule*> incrulevect_t;

// class definition
class TScriptRule
{
    // Hold tokens in separate rule
    AtvfRule*       _atvfToknRuleP;
    // Contains a rule for each include file
    incrulevect_t   _atvfRulesVect;
    // Provide a rule look up
    toknrulemap_t   _objRuleMap;
    //  --------------------------------------------------------
    TRecCreator* _headCreateP;
    trec_crtrs_t _creators;
    keyvect_t _creatorKeyIdx; // To save ODl ordering

    // ODL atvf rule moved to attributes so it may be shared when reading all included
    // ODL files.
    static       AtvfRule        _odl;       // ODL rule
    static const AtvfSimplRule   _prmRule;   // Parameter rule
    static const AtvfSimplRule   _recRule;   // Record rule
    static const AtvfSimplRule   _optRule;   // Option rule
    static const AtvfDelimRule   _lopRule;   // Long option rule
    static const AtvfSimplRule   _defRule;   // Default rule
    static const AtvfSimplRule   _atrRule;   // Attribute rule
    static const AtvfSimplRule   _rowRule;   // Table row rule
    static const AtvfSimplRule   _colRule;   // Table column rule
    static const AtvfSimplRule   _incRule;   // Include file rule
    // Causes rule to created.
    static const bool _odlRuleCreated;
    static const bool _createOdlRule();

    TParamRule* makePrmRule(const string& text, prulep_cont_t& params);
    // Helper recursive read odl function.
    void read_odl(const string& odlname, bool include_odl = false);
public:
    //Constructors & Destructor
    TScriptRule(const string& odlname = "atvf.odl");
    ~TScriptRule();

    //Accessors
    TScriptHead* extractHead(AtvfInput& in) const;
    TScriptHead* newHead() const;
    TRecord* extractRecord(AtvfInput& in) const;
    keyvect_t* getRecordsKeys() const;
    strvect_t* getRecordsInfo(const string& optkey,
                              const keyvect_t* kVectP = NULL) const;
    TRecord* makeNewRecord(const AtvfKey& key) const;
    void saveHead(const string& version, ostream& out) const;
    const TParamRule* getParamRule(const string& ruleName,
                                   const string& creatName = "head" ) const;

    //Mutators

    //Friend Operators
    friend ostream& operator <<(ostream& os, TScriptRule& TS);
private:
    TScriptRule(const TScriptRule&);
    operator=(const TScriptRule&);
}
; // TScriptRule

#endif // _CYR_TSCRIPTRULE_H
