/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/


#include "mdbparse.h"
#include "mdbstl.h"

extern "C"
{
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
}

const string Cmdbparse::wordBreak = " \n\r\t";
#define CHECK_END(s, p) if(p >= s.length()) return -1;
#define SKIP_WHITESPACE(s, p)   if(sw) \
{\
while(isspace(s[p])) \
{\
p++;\
CHECK_END(s, p)\
}\
}


// Cmdbparse implementation ////////////////////////////////////////////

Cmdbparse::Cmdbparse()
        : clp(0), sw(false), cs(false)
{}


Cmdbparse::Cmdbparse(const string & cline)
        : cl(cline), clp(0), sw(false), cs(false)
{}


Cmdbparse::~Cmdbparse()
{}


void Cmdbparse::set_text(string & cline)
{
    cl = cline;
    clp = 0;
}

string & Cmdbparse::get_text(void)
{
    return cl;
}

// Enable/disable case sensitive character and word compares
void Cmdbparse::case_sensitive(bool y)
{
    cs = y;
}

// Skip over whitespace (space tabs etc) when parsing line
void Cmdbparse::skip_whitespace(bool y)
{
    sw = y;
}

// Remove from line any continuation characters listed in "chars"
// if they preceed a newline or carriage return or at the
// end of the string
void Cmdbparse::remove_contin_chars(const string & chars)
{
    string::size_type i = 0;
    string::size_type p;

    while(1)
    {
        // Find first continuation character, return if none
        if((i = cl.find_first_of(chars, i)) == cl.npos)
            return;

        // Find first non space or tab after continuation
        p = cl.find_first_not_of(" \t", i+1);
        // If end of string, continuation shouldn't be there
        if(p == cl.npos)
        {
            // Replace continuation with space
            cl.replace(i, 1, " ");
            // Done
            return;
        }
        // If newline or return, we found a real continuation char
        else if( (cl[p] == '\n') || (cl[p] == '\r') )
        {
            // Replace continuation with space
            cl.replace(i, 1, " ");
            // Update continuation char search start pos
            i = p;
        }
        else
        {
            i = p;
        }
    }
}

// Replace all instances of character "oldchar" with "newchar"
int Cmdbparse::replace_char(const char oldchar, const char newchar)
{
    string::size_type p = 0;
    while((p = cl.find_first_of(oldchar, p)) != cl.npos)
        // Replace continuation with space
        cl.replace(p++, 1, 1, newchar);
    return 0;
}

// Expect specified character sequence
int Cmdbparse::expect(const string & word)
{
    // Input word iterator
    string::size_type   iw = 0;
    // Temporary command line iterator
    string::size_type   ci = clp;

    // Skip leading whitespace if option set
    SKIP_WHITESPACE(cl, ci)

    // Loop until end of input word
    while(iw < word.length())
    {
        // Case sensitive comparison
        if(cs)
        {
            // Return zero if non matching character found
            if( cl[ci++] != word[iw++] )
                return 0;
        }
        // Non case sensitive
        else
        {
            // Return zero if non matching character found
            // after converting both to same case
            if( toupper(cl[ci++]) != toupper(word[iw++]) )
                return 0;
        }
    }
    // Move command line pointer (iterator) to next character
    clp = ci;
    // Return found
    return 1;
}

// Check if next character matches "c"
int Cmdbparse::expect(const char c)
{
    // Temporary command line iterator
    string::size_type ci = clp;

    // Skip leading whitespace if option set
    SKIP_WHITESPACE(cl, ci)

    // Case sensitive comparison
    if(cs)
    {
        if( cl[ci] != c )
        {
            // No match
            return 0;
        }
    }
    // Non case sensitive
    else
    {
        if( toupper(cl[ci]) != toupper(c) )
        {
            // No match
            return 0;
        }
    }
    // Move command line pointer (iterator) to next character
    clp = ++ci;
    // Return found
    return 1;
}

// Check if next character is one of those in "chars"
// Return the (1 based) index in "chars" of the match or
// zero if none found.
int Cmdbparse::expect_one_of(const string & chars)
{
    // Temporary command line iterator
    string::size_type ci = clp;
    // Input character list iterator
    string::size_type iw = 0;
    // Skip leading whitespace if option set
    SKIP_WHITESPACE(cl, ci)

    // Loop through "chars" to see if next command line
    // character matches one of those expected.
    for(int i = 1 ; i <= chars.length() ; i++)
    {
        // Case sensitive comparison
        if(cs)
        {
            // Match found ?
            if( cl[ci] == chars[iw++] )
            {
                // Move command line pointer
                clp = ++ci;
                // Return matching index in "chars"
                return i;
            }
        }
        // Non case sensitive
        else
        {
            // Match found ?
            if( toupper(cl[ci]) == toupper(chars[iw++]) )
            {
                // Move command line pointer
                clp = ++ci;
                // Return matching index in "chars"
                return i;
            }
        }
    }
    // Not found
    return 0;
}

// Return next character in line without moving pointer
char Cmdbparse::peek_next_char(void)
{
    string::size_type ci = clp;
    // Skip leading whitespace if option set
    SKIP_WHITESPACE(cl, ci)
    // Return the character
    return cl[ci];
}

// Return next character in line
char Cmdbparse::get_next_char(void)
{
    // Skip leading whitespace if option set
    SKIP_WHITESPACE(cl, clp)
    // Return the character, move pointer to next
    return cl[clp++];
}

// Get text from line until a character from "uchars" is found.
// Return the (1 based) index of the character from "uchars"
// which terminated the get.
int Cmdbparse::get_text_until(string & opword, const string & uchars)
{
    // Terminator character list iterator
    string::size_type iw;

    // Skip leading whitespace if option set
    SKIP_WHITESPACE(cl, clp)
    // Ensure output buffer is empty
    BSCLEAR(opword);

    while( clp < cl.length() )
    {
        // Loop through "uchars" to see if next command line
        // character matches one of the terminators
        for(iw = 0 ; iw < uchars.length() ; iw++)
        {
            // Case sensitive comparison
            if(cs)
            {
                // Match found ?
                if( cl[clp] == uchars[iw] )
                {
                    // Return matching index in "chars"
                    return iw+1;
                }
            }
            // Non case sensitive
            else
            {
                // Match found ?
                if( toupper(cl[clp]) == toupper(uchars[iw]) )
                {
                    // Return matching index in "chars"
                    return iw+1;
                }
            }
        }
        // Append character to output word
        opword += cl[clp++];
    }
    // Return end of input string
    return -1;
}

// Get integer value from line. Output is provided in
// integer and string form. Return value is zero on error
// -1 on end of input.
int Cmdbparse::get_integer(int & opint, string & opstr)
{
    // Ensure output string is empty
    BSCLEAR(opstr);
    opint = 0;


    CHECK_END(cl, clp)

    // Skip leading whitespace if option set
    SKIP_WHITESPACE(cl, clp)

    // Check first character
    // Allow leading plus sign
    if( cl[clp] == '+' )
    {
        // Skip over it
        clp++;
    }
    // Check for leading minus sign
    else if( cl[clp] == '-' )
    {
        // Append it to output string
        opstr += cl[clp++];
    }

    CHECK_END(cl, clp)

    // Get position of next non numeric digit
    string::size_type i = cl.find_first_not_of("0123456789", clp);
    // Return if no digits at current index
    if(i == clp)
        return 0;

    // Integer may extend to end of string
    else if(i == cl.npos)
    {
        // Append digits to output string
        opstr += cl.substr(clp);
        // Move pointer beyond end
        clp = cl.length();
    }
    else
    {
        // Append digits to output string
        opstr += cl.substr(clp, i - clp);
        // Move pointer after digits
        clp = i;
    }
    // Create integer version of output string.
    sscanf(opstr.c_str(), "%d", &opint);
    return 1;
}

// Get string value from line delimited by character "delim".
// The delimiters are removed from the output string.
int Cmdbparse::get_delim_string(string & str, char delim)
{
    string::size_type ci = clp;

    // Skip leading whitespace if option set
    SKIP_WHITESPACE(cl, ci)

    // Ensure first character is a delimiter
    if( cl[ci] != delim )
        return 0;

    // Get position of next delimiter
    string::size_type i = cl.find_first_of(delim, ++ci);
    // Return error if no end delimiter found
    if( i == cl.npos )
        return 0;

    // Append digits to output string
    str = cl.substr(ci, i - ci);
    // Move pointer after string end delimiter
    clp = i+1;

    return 1;
}

