/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_ASYNCSECURESOCKET_H_
#define _OPENSTA_ASYNCSECURESOCKET_H_

////////////////////////////////////////////////////////////////////////////////
// Asynchronous secure socket definition file.
// This class implements a secure communication channel using a transport based
// on an asynchronous socket (CAsyncTCPSocket).
//
// The supported secure protocols are:
// 1. PCT
// 2. SSL 2.0
// 3. SSL 3.0
// 4. TLS 1.0
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

// Force linking against the Crypt32 library
#pragma comment(lib, "Crypt32.lib")

#include "AsyncTCPSocket.hpp"
#include <schannel.h>
#define SECURITY_WIN32
#include <security.h>
#include <sspi.h>
#include <map>
#include "FastLock.hpp"


// This class implements an asynchronous secure socket bound to an I/O completion port.
// THREAD UNSAFE. This should not be a problem, since an asynchronous device should
// always be accessed by a single thread at a time. The derivatives should be thread
// unsafe as well.
class CAsyncSecureSocket : public CAsyncTCPSocket, public IocpEvent
{
public:
    enum EventTypes
    {
        EVENT_TYPE_CONNECTION_ESTABLISHED               = 0x10,
        EVENT_TYPE_MESSAGE_SENT                         = 0x11,
        EVENT_TYPE_MESSAGE_RECEIVED                     = 0x12,
        EVENT_TYPE_CONNECTION_CLOSED                    = 0x13,
        EVENT_TYPE_PLEASE_PROVIDE_CLIENT_CERTIFICATE    = 0x14,
    };

    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------

    // Constructor.
    // To set the read buffer for the instance, pass a non-null pbReadBuffer, and
    // pass its size in bytes as dwSize. This buffer is invisible to the user of
    // the instance, so calling this method to provide it is purely a question of
    // optimization. However, if the buffer you supply is not large enough, the
    // Open, Read, Write and Close methods can return an error. Call the static
    // function GetReadBufferMinSize() to determine the minimum size to allocate.
    // By default, the instance allocates the read buffer during Open and frees it
    // in Close.
    CAsyncSecureSocket(CIOCP& iocp, PBYTE pbReadBuffer = NULL, DWORD dwSize = 0, CTrace* pTrace = NULL);

    // Destructor.
    virtual ~CAsyncSecureSocket();


    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------

    // Initialize the SSPI. This does NOT call the base class's Init() method.
    // MUST be called AFTER CFastLock.Init();
    static DWORD Init();

    // Uninitialize the SSPI. This does NOT call the base class's Uninit() method.
    static void Uninit()
    {
        // Free the anonymous credentials handle.
        m_securityFuncTbl.FreeCredentialsHandle(&m_hAnonymCreds);
        ZeroMemory(&m_hAnonymCreds, sizeof(m_hAnonymCreds));

        // Free the credentials handles stored in the cache, and empty the cache.
        for(CertificateToHandle::iterator pCert = m_credentialsCache.begin(); pCert != m_credentialsCache.end(); ++pCert)
            m_securityFuncTbl.FreeCredentialsHandle(&pCert->second);
        m_credentialsCache.clear();

        // Free the security DLL and functions table.
        ZeroMemory(&m_securityFuncTbl, sizeof(m_securityFuncTbl));
        if(m_hSecurDll)
        {
            FreeLibrary(m_hSecurDll);
            m_hSecurDll = NULL;
        }

        // Delete the credentials handle cache lock.
        if(m_pLock)
        {
            delete m_pLock;
            m_pLock = NULL;
        }
    }

    // Set the common name of the client certificate to be used when requested
    // by server. NULL = anonymous user.
    void SetClientCertificateName(LPCTSTR zClientCertificate = NULL)
    {
        if(zClientCertificate == NULL)
            m_zCertifName[0] = 0;
        else
            lstrcpy(m_zCertifName, zClientCertificate);
    }

    // Sets a new read buffer.
    // This function must NOT be called while the secure socket is open.
    // It allows to change the internal read buffer after constructing an instance,
    // an change it again later. The minimum size of the buffer you pass in
    // should be determined using GetReadBufferMinSize().
    // The function returns the previous read buffer in the pbPrevious variable
    // if it had not been auto-allocated, that is, if it was passed to the
    // constructor or set by a previous call to this function.
    // To re-enable auto-allocation, call the function with a NULL read buffer.
    DWORD SetReadBuffer(PBYTE pbReadBuffer, DWORD dwSize, PBYTE& pbPrevious)
    {
        DWORD dwRet = 0;
        assert(!IsOpen());
        pbPrevious = m_bAutoAlloc ? NULL : m_pReadBuffer;
        if(!m_bAutoAlloc || pbReadBuffer != NULL)   // No use passing NULL buffer when already auto-alloc.
        {
            if(m_bAutoAlloc)
                delete [] m_pReadBuffer;
            m_pReadBuffer  = pbReadBuffer;
            m_dwReadBufLen = dwSize;
        }
        return dwRet;
    }

    // Get the minimum size of the internal read buffer, in bytes.
    // The returned size is rounded up to a multiple of DWORD.
    static DWORD GetReadBufferMinSize(DWORD& dwSize)
    {
        DWORD dwRet = 0;
        PSecPkgInfo psecInfo;
        SECURITY_STATUS ss = m_securityFuncTbl.QuerySecurityPackageInfo(UNISP_NAME, &psecInfo);
        if(ss == SEC_E_OK)
        {
            dwSize = psecInfo->cbMaxToken + 1024;   // + 1024 is empirical: it seems that
                                                    // the raw psecInfo->cbMaxToken value is
                                                    // sometimes too small?!
            // Round up to a DWORD multiple.
            if((dwSize & 0x00000003) != 0)
                dwSize = (dwSize & ~0x00000003) + 4;
            m_securityFuncTbl.FreeContextBuffer(psecInfo);
        }
        else
            dwRet = (DWORD)ss;
        return dwRet;
    }

    // Create and connect the socket, then issue a client handshake for
    // an anonymous connection. If the socket is already connected,
    // then only the handshake is performed. If a client certificate
    // name was supplied via SetClientCertificateName(), then the
    // corresponding certificate is used in the handshake.
    // zServer can be a name or IP address.
    // The socket is then bound to the IOCP.
    // If zServer if an IP address syntax, no name resolution is used.
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Open(LPCSTR zServer, DWORD dwPort, IocpEvent *pEventTarget);

    // Disconnect and close the socket.
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Close();

    // Asynchronous socket receive and message decrypt operation.
    // If zBuffer is NULL, then it means the caller is not interested in the results.
    // CAUTION: the buffer must remain valid until the call completes!
    // Return value: 0 = success, other = WSAGetLastError() failure code.
    virtual DWORD Read(LPSTR zBuffer, DWORD dwBufferLength);
    DWORD Read(LPWSABUF lpBuffers, DWORD dwBufferCount);

    // Message encrypt and asynchronous socket send operation.
    // If dwBufferLength is 0, then zBuffer is considered NULL-terminated.
    // CAUTION: the buffer must remain valid until the call completes!
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Write(LPCSTR zBuffer, DWORD dwBufferLength = 0);
    DWORD Write(LPWSABUF lpBuffers, DWORD dwBufferCount);

    // Is the device open?
    virtual BOOL IsOpen();

    //// --------------------------------------------------------------------------------
    //// IocpEvent methods.
    //// --------------------------------------------------------------------------------
    virtual void OnEvent      (DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType);
    virtual void OnEventFailed(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType, DWORD dwError);

protected:
    //// --------------------------------------------------------------------------------
    //// Types.
    //// --------------------------------------------------------------------------------
    enum State
    {
        STATE_NONE,
        STATE_INITIAL_BLOB_SENT,
        STATE_HANDSHAKE_DATA_RECEIVED,
        STATE_HANDSHAKE_DATA_SENT,
        STATE_CLOSE_DATA_SENT,
        STATE_DATA_SENT,
        STATE_CLOSE_ACKNOWLEDGE_DATA_SENT,
        STATE_DATA_RECEIVED,
    };

    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------
    void  Reset();          // Clears data members so that instance is ready to use.
    DWORD InitiateClientHandshakeLoop(BOOL bRenegotiate = FALSE);
    DWORD ClientHandshakeLoop();
    DWORD ReadHandshakeData();
    DWORD WriteHandshakeData(DWORD dwNextState);
    DWORD WriteData(DWORD dwState = STATE_DATA_SENT);
    SECURITY_STATUS ProcessBlob(DWORD& dwSSPIOutFlags);
    void  MoveExtraHandshakeData();
    void  ProcessSuccessfulHandshake();
    BOOL  MakeClientCertificateCredentials();
    DWORD CompleteClose(BOOL bNotify = TRUE);
    DWORD CompleteCloseAcknowledge();
    DWORD ReadLoop();
    void  CopyToClientBuffers(PVOID pvSource, DWORD dwLen);
    DWORD CopyFromClientBuffers(LPWSABUF lpSource, DWORD dwBufferCount, PVOID pvTarget, DWORD dwLen);
    std::ostream& TraceHeader(std::ostream& os);
    DWORD AllocReadBuffer() // Allocate the read buffer.
    {
        DWORD dwRes = 0, dwMaxToken = 0;
        GetReadBufferMinSize(dwMaxToken);
        if(m_pReadBuffer)
        {
            // Read buffer has already been inherited.
            // Just check the buffer's length.
            if(m_dwReadBufLen < dwMaxToken)
                dwRes = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            // Allocate a new read buffer.
            m_dwReadBufLen = dwMaxToken;
            m_pReadBuffer = new BYTE[m_dwReadBufLen];
            if(m_pReadBuffer)
                m_bAutoAlloc = TRUE;
            else
                dwRes = ERROR_NOT_ENOUGH_MEMORY;
        }
        return dwRes;
    }
    void FreeReadBuffer()   // Free the read buffer.
    {
        if(m_bAutoAlloc)
        {
            if(m_pReadBuffer)
            {
                delete [] m_pReadBuffer;
                m_pReadBuffer = NULL;
            }
            m_bAutoAlloc = FALSE;
        }
    }


    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------
    static HMODULE                m_hSecurDll;        // The security DLL: secur32.dll (Windows 2000); security.dll (Windows NT 4.0).
    static SecurityFunctionTable  m_securityFuncTbl;  // The security function table.
    static BOOL                   m_bRunAsSystem;     // TRUE if the code is being executed under the LocalSystem account.
    static CredHandle             m_hAnonymCreds;     // Anonymous credentials handle.
    static CFastLock*             m_pLock;            // Lock for the credentials cache.
    typedef std::map<std::string,CredHandle> CertificateToHandle;
    static CertificateToHandle    m_credentialsCache; // The credentials cache.
    TCHAR                         m_zCertifName[512]; // The name of the client certificate to use if needed.
    CredHandle                    m_hCredentials;     // The current credentials handle.
    CtxtHandle                    m_hContext;         // Context handle.
    IocpEvent*                    m_pEventTarget;     // The event target passed to Open().
    DWORD                         m_dwState;          // Carrying the state across asynchronous calls.
    SecBuffer                     m_buf1[1], m_buf2[4]; // Temporary buffers used during client handshake.
    SECURITY_STATUS               m_ss;               // Status during client handshake loop.
    BOOL                          m_bDoRead;          // Used during client handshake loop.
    BOOL                          m_bRenegotiate;     // TRUE if a renegotiation is in progress.
    BOOL                          m_bConnected;       // TRUE when a successful client handshake has been performed.
    PBYTE                         m_pReadBuffer;      // Buffer allocated for incoming messages.
    DWORD                         m_dwReadBufLen;     // Length of m_pReadBuffer.
    DWORD                         m_dwBytesReceived;  // Number of bytes received so far during a read.
    BOOL                          m_bAutoAlloc;       // TRUE if m_pReadBuffer was allocated by the instance.
    LPWSABUF                      m_pClientBuffer;    // Client buffers to store received data.
    DWORD                         m_dwBufferCount;    // Number of client buffers.
    DWORD                         m_dwTotalBufLen;    // Cumulated length of client buffers.
    WSABUF                        m_oneBuf;           // Unique buffer used by Read(LPSTR,DWORD).
    SecBuffer*                    m_pDataBuffer;      // Temporary pointer to read data.
    SecBuffer*                    m_pExtraBuffer;     // Temporary pointer to extra data.
    DWORD                         m_dwNextChunk;      // Offset of the next chunk of data available to the caller of Read.

private:

    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    void operator=(const CAsyncSecureSocket&);
};

#endif  // _OPENSTA_ASYNCSECURESOCKET_H_
