/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "stdafx.h"
#include "AsyncVirtualUserController.h"
#include "CORBARepository.h"

#ifndef CYRANO_TP_STANDALONE
#   include "CECSemaphore.h"
#endif

#include "TEMainStructure.hpp"

extern CMainExecuterStructure *g_pTExecuter;   // Defined and initialised in main.cpp.

extern CORBAServer*    pCServer;
extern CORBARepository* Repository;

CFastLock *CAsyncVirtualUser::s_pCacheSl = NULL;

// ----------------------------------------------------------------------------
//
//  CScriptContextData
//
// ----------------------------------------------------------------------------

CScriptContextData::CScriptContextData(DWORD dwParallelism)
{
    m_pTaskScript           = NULL;
    m_dwTRecordCallStep     = 0;
}


// ----------------------------------------------------------------------------
//
//  CAsyncVirtualUser
//
// ----------------------------------------------------------------------------

CAsyncVirtualUser::CAsyncVirtualUser()
    : m_pTimerQueue(NULL), m_dwTestCaseFail(0), m_dwTestCasePass(0)
#ifndef CYRANO_TP_STANDALONE
    , m_pCurrentSemaphore(NULL)
#endif
{
    assert(s_pCacheSl != NULL);
}

CAsyncVirtualUser::~CAsyncVirtualUser()
{
#ifndef CYRANO_TP_STANDALONE
    if(m_pCurrentSemaphore)
        delete m_pCurrentSemaphore;
#endif
}

DWORD CAsyncVirtualUser::Init(CThreadPoolController& controller, CTimerQueue* pTimerQueue, DWORD dwManagerNumber, DWORD dwParallelism, CTrace* pTrace)
{
    DWORD dwError = 0;

    m_dwParallelism     = dwParallelism;
    m_execPhase         = EXEC_INIT;

    m_pTimerQueue = pTimerQueue;

    dwError = CAsyncHTTPRequestManager::Init(controller, dwManagerNumber, dwParallelism, pTrace);

    return dwError;
}

BOOL CAsyncVirtualUser::InitClass()
{
    assert(s_pCacheSl == NULL);

    if(s_pCacheSl == NULL)
        s_pCacheSl = new CFastLock();

    if(s_pCacheSl == NULL)
        return FALSE;
    else
        return TRUE;
}

void CAsyncVirtualUser::UninitClass()
{
    if(s_pCacheSl != NULL)
    {
        delete s_pCacheSl;
        s_pCacheSl = NULL;
    }
}

// Mainly used to tell the VU what TOF file it's gonna use.
// szTofFile: Full path, name and extension of the file.
void CAsyncVirtualUser::Initialise(const string &szTofFile
#ifndef CYRANO_TP_STANDALONE
                                   , const string &szTestName, const ThreadDef &thrDef
#endif
                                   )
{
#ifndef CYRANO_TP_STANDALONE
    m_szTestName    = szTestName;   // TODO: get_test_name might be called even in standalone mode. Is that a pb?
    m_ThreadDef     = thrDef;
#endif

    m_szMainTOFFile = szTofFile;

    // Derives from TContext. Must call InitContextData().
    InitContextData();
    assert(CURR_SCRIPT_CTXT_DATA != NULL);

    set_state(MDB_SRSIG);   // Into TContext.
}

// Run thread task(s)
// Can be called by kernel thread start routine, must be mt safe
// Code got inspired from "int Cmdbthread::run(void)" and "ThreadTask::run"

// Executes TOF sequences. Ends when the TOF needs to wait for an IO or sync. op.
// Gets called again when IO is over.
// Handles the recursive call of scripts, finishing the calling scripts execution when
// the called script is over.
// Used both internally and by CAsyncSecureSocket in case SSL Connect/Disconnect are executed.

ETofResult CAsyncVirtualUser::ProcessTof()
{
    if(GetTrace() && GetTrace()->IsTrace("full"))
    {
        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::ProcessTof: thread " << GetCurrentThreadId() << " processing user " << m_dwManagerNumber << ".\r\n";
        ReleaseTraceStream();
    }

    // Process current thread state.
    m_execPhase = EXEC_RUNNING;
    ETofResult res;
    BOOL bContinue = TRUE;
    
    do
    {
        res = CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetScript()->run(this);

        switch(res)
        {
        case TOF_RES_FAILED :
            {
                // Aborting.
                m_execPhase = EXEC_OVER;

                // Signal we are over.
                CleanUp();  // Cf. TContext or derived.

                ((CAsyncVirtualUserController*)m_pController)->IncFailedUsers();
                m_pController->OnAllRequestsFinished(m_dwManagerNumber);

                return TOF_RES_FAILED;
            }

        case TOF_RES_ASYNC_STARTED :
            {
                // IO under process. Let the framework do its job and call us again.
                return TOF_RES_ASYNC_STARTED;
            }

        case TOF_RES_OVER :
            {
                // End-of-Executable-Records for the current thread. Remaining records are procedure definitions.
                // If this is the base-script, then signal the controller we are over with our TOF list.

                m_execPhase = EXEC_OVER;

                // End TOF processing.
                if(m_ContextDataStack.size() == 1)
                {
                    // End of that VU simulation.
                    ((CAsyncVirtualUserController*)m_pController)->IncSuccessfullUsers();
                    bContinue = FALSE;
                }
                else
                {
                    // End of current script. Will continue with the calling one.
                }

                EndScript();

                break;
            }

        case TOF_RES_KILL_THREAD :
            {
                // Must terminate all scripts.
                while(m_ContextDataStack.size() > 1)
                {
                    EndScript();
                }

                // End of that VU simulation.
                ((CAsyncVirtualUserController*)m_pController)->IncSuccessfullUsers();

                bContinue = FALSE;

                EndScript();

                break;
            }

/*
        case TOF_RES_OK :
        case TOF_SKIP_TO_NEXT :
            {
                // Never used or error since processed in TScript::run().
            }
*/
        default :
            {
                // Error: should have been processed in TScript::run.
                assert(0);
                break;
            }
        };
    }
    while(bContinue == TRUE);

    return res;
}


void CAsyncVirtualUser::OnRequestFinished(DWORD dwRequestSlotNumber)
{
    ProcessEndOfRequest(dwRequestSlotNumber, FALSE);
}

void CAsyncVirtualUser::OnRequestCancelled(DWORD dwRequestSlotNumber)
{
    ProcessEndOfRequest(dwRequestSlotNumber, TRUE);
}

// --------------------------------------------------------------------------------
// Called when a request is over.
// If a synchronisation operation was initiated then the TOF processing for that
// VU was waiting for us, thus must enter ProcessTof again.
// Otherwise, the TOF list is being processed in an other thread.
// --------------------------------------------------------------------------------
void CAsyncVirtualUser::ProcessEndOfRequest(DWORD dwRequestSlotNumber, BOOL bWasCancelled)
{
    assert(dwRequestSlotNumber != ULONG_MAX);

    DWORD           dwLocPendingRequests;
    EVuExecPhase    locExecPhase;

    {
        CLock guard (m_sl);

        // Code based on CAsyncRequestManager::OnRequestFinished.
        --m_dwPendingRequests; // Code based on CAsyncRequestManager::OnRequestFinished.

        string zTimerName = ((CAsyncHTTPRequest*)m_requests[dwRequestSlotNumber])->GetTimerName();
        if(!zTimerName.empty())
            StopTimer(zTimerName);

        dwLocPendingRequests = m_dwPendingRequests;

        locExecPhase = m_execPhase;
    }

    // Handle synchronisation requests, and move further.
    switch(locExecPhase)
    {
    case EXEC_HTTP_SYNC :
        {
            // Last TOF object required to wait for a specific HTTP request to complete.
            if(((CAsyncHTTPRequest*)m_requests[dwRequestSlotNumber])->GetConnectionId() == m_dwWaitedConnId)
            {
                // This is THE request we were waiting for. Get back to work. Usually, will process the current request's results.
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": connection " << m_dwManagerNumber << '/' << m_dwWaitedConnId << " is over and was waited upon.\r\n";
                    ReleaseTraceStream();
                }

                ProcessTof();   // No need to handle the results.
            }
            else
            {
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": user " << m_dwManagerNumber << " waiting for " << m_dwWaitedConnId << ". " << dwLocPendingRequests << " pending.\r\n";
                    ReleaseTraceStream();
                }
            }
            // Otherwise, don't care for that request's result.
            break;
        }

    case EXEC_ALL_HTTP_SYNC :
        {
            // Last TOF object required to wait for all the HTTP requests to complete.
            if(dwLocPendingRequests == 0)
            {
                // This was the last executing request in the list.
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": user " << m_dwManagerNumber << " releasing the sync. mechanism.\r\n";
                    ReleaseTraceStream();
                }

                ProcessTof();   // No need to handle the results.
            }
            else
            {
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": user " << m_dwManagerNumber << " waiting for " << dwLocPendingRequests << " more requests to complete.\r\n";
                    ReleaseTraceStream();
                }
            }
            break;
        }

    default :
        {
            // Nothing particular to do.
            if(GetTrace() && GetTrace()->IsTrace("full"))
            {
                GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": connection " << m_dwManagerNumber << '/' << ((CAsyncHTTPRequest*)m_requests[dwRequestSlotNumber])->GetConnectionId() << " is over.\r\n";
                ReleaseTraceStream();
            }
            break;
        }
    };
}

void CAsyncVirtualUser::OnEvent(DWORD /*dwNumberOfBytesTransferred*/, BOOL bIsIo, DWORD dwEventType)
{
    if(bIsIo == TRUE)
    {
        if(GetTrace() && GetTrace()->IsTrace("errors"))
        {
            GetTraceStream(m_pTrace) << "CAsyncVirtualUser::OnEvent: ERROR. Client " << m_dwManagerNumber << " received an unexpected IO.\r\n";
            ReleaseTraceStream();
        }
    }
    else
    {
        switch(dwEventType)
        {
        case CAsyncVirtualUser::EVENT_TYPE_VIRTUAL_USER_START:
            {
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEvent: client " << m_dwManagerNumber << " is starting.\r\n";
                    ReleaseTraceStream();
                }

                OnStartProcessing();
                break;
            }

        case IocpEvent::EVENT_TYPE_TIMER:
            {
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEvent: client " << m_dwManagerNumber << " received a timer event.\r\n";
                    ReleaseTraceStream();
                }

                OnTimerFinished();
                break;
            }

        default:
            if(GetTrace() && GetTrace()->IsTrace("errors"))
            {
                GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEvent: client " << m_dwManagerNumber << " received an unsupported event: " << dwEventType << ".\r\n";
                ReleaseTraceStream();
            }
        };
    }
}

void CAsyncVirtualUser::OnEventFailed(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType, DWORD dwError)
{
    if(GetTrace() && GetTrace()->IsTrace("errors"))
    {
        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEventFailed: ERROR. Client " << m_dwManagerNumber << " failed.\r\n";
        ReleaseTraceStream();
    }

    if(bIsIo == TRUE)
    {
        if(GetTrace() && GetTrace()->IsTrace("errors"))
        {
            GetTraceStream(m_pTrace) << "CAsyncVirtualUser::OnEventFailed: ERROR. Client " << m_dwManagerNumber << " received an unexpected IO.\r\n";
            ReleaseTraceStream();
        }
    }
    else
    {
        // Execute despite an error was received.
        switch(dwEventType)
        {
            case IocpEvent::EVENT_TYPE_TIMER:
                OnTimerFinished();
                break;

            default:
                if(GetTrace() && GetTrace()->IsTrace("errors"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEventFailed: client " << m_dwManagerNumber << " received an unsupported event: " << dwEventType << ".\r\n";
                    ReleaseTraceStream();
                }
        };
    }
}

void CAsyncVirtualUser::OnStartProcessing()
{
    if(GetTrace() && GetTrace()->IsTrace("Startup"))
    {
        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnStartProcessing: client " << m_dwManagerNumber << " is starting.\r\n";
        ReleaseTraceStream();
    }

    m_execPhase = EXEC_RUNNING;
    time_t t;
    time(&t);
    m_dwStartTime = t;

    StartScript(m_szMainTOFFile);
}

void CAsyncVirtualUser::OnTimerFinished()
{
    if(m_execPhase == EXEC_WAIT_TIMER)
    {
        ProcessTof();   // Nb: results handled in ProcessTof().
    }
    else
    {
        // Strange: How come we were signaled?
        assert(0);
    }
}

void CAsyncVirtualUser::OnTimerCancelled()
{
    if(m_execPhase == EXEC_WAIT_TIMER)
    {
        ProcessTof();   // Nb: results handled in ProcessTof().
    }
    else
    {
        // Strange: How come we were signaled?
        assert(0);
    }
}

// --------------------------------------------------------------------------------
// An async (SSL) disconnect happened. Must resume TOF processing.
// --------------------------------------------------------------------------------
void CAsyncVirtualUser::OnRequestClosed(DWORD dwRequestNumber)
{
    CLock lock(m_sl);

    // Call base class' implementation.
    CAsyncRequestManager::OnRequestClosed(dwRequestNumber);

    assert(m_dwDisconnectToWait>0);
    --m_dwDisconnectToWait;

    if(m_execPhase == EXEC_WAIT_DISCONNECTS)
    {
        if(m_dwDisconnectToWait == 0)
        {
            if(GetTrace() && GetTrace()->IsTrace("connect sync"))
            {
                GetTraceStream(GetTrace()) << "EXEC WAIT DISCONNECTS: all connections disconnected.\r\n";
                ReleaseTraceStream();
            }

            ProcessTof();   // No need to handle the results.
        }
    }
}


///////////////////////////////////////////////////////////////////////////////
////
////  Code which purpose is HTTP request slot selection / allocation / connection.
////
///////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------------
// GetRequestFromConnId
//
// As connection IDs are incremented each time a new connection is established,
// that method relates a connection ID to the connection slot it was assigned at
// connection time.
//
// Protected method. Serialization must be handled by calling methods.
// --------------------------------------------------------------------------------
CAsyncHTTPRequest* CAsyncVirtualUser::GetRequestFromConnId(DWORD dwConnectionId) const
{
    assert(dwConnectionId != ULONG_MAX);

    // Search for the right slot.
    DWORD dwSize = m_requests.size();
    for(DWORD dwIndice=0 ; dwIndice<dwSize ; ++dwIndice)
    {
        if(((CAsyncHTTPRequest*)m_requests[dwIndice])->GetConnectionId() == dwConnectionId)
        {
            return (CAsyncHTTPRequest*)m_requests[dwIndice];
        }
    }

    return NULL;
}

// --------------------------------------------------------------------------------
// GetNotConnectedRequest
//
// Find a free request. If none is available, then return NULL.
// Protected method. Serialization must be handled by calling methods.
// --------------------------------------------------------------------------------
CAsyncRequest* CAsyncVirtualUser::GetNotConnectedRequest()
{
    // Return the first request found whose processing is over and for which results are no more used.
    DWORD dwReqId = CAsyncRequestManager::GetAvailableRequest();

    if(dwReqId == ULONG_MAX)
    {
        // Allocate a new request. Would not be used if TOF was recorded from an HTTP-compliant browser (max 4 connections).
        CAsyncHTTPRequest* pNewReq = new CAsyncHTTPRequest(m_pController->GetIocp(), *this, m_requests.size(), m_pTrace);
        pNewReq->Init();

        AddRequest(pNewReq);

        return pNewReq;
    }
    else
    {
        return m_requests[dwReqId];
    }
}

// --------------------------------------------------------------------------------
// ObtainRequestStruct
//
// Retrieve or allocate a request slot and assign it to connection ID dwConnectionId. 
// Called by TRcrdConnect and TRcrdRequest.
// --------------------------------------------------------------------------------
CAsyncHTTPRequest* CAsyncVirtualUser::ObtainRequestStruct(DWORD dwConnectionId)
{
    CAsyncHTTPRequest* pNewReq;

    {
        CLock lock(m_sl);

        // First check if that connection has already been allocated.
        pNewReq = GetRequestFromConnId(dwConnectionId);
        if(pNewReq == NULL)
        {
            // No connection allocated for that connection ID. Allocate a new one.
            pNewReq = (CAsyncHTTPRequest*)GetNotConnectedRequest();
            if(pNewReq != NULL)
            {
                pNewReq->SetConnectionId(dwConnectionId);
            }
            else
            {
                // Unexpected error.
                assert(0);
                return NULL;
            }
        }
    }

    if(GetTrace() && GetTrace()->IsTrace("connection"))
    {
        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::ObtainRequestStruct associated request slot " << pNewReq->GetRequestNumber() << " to connection ID " << dwConnectionId << "\r\n";
        ReleaseTraceStream();
    }

    return pNewReq;
}

// --------------------------------------------------------------------------------
// ReleaseRequest
//
// Release the connection slot associated with dwConnectionId.
// --------------------------------------------------------------------------------
void CAsyncVirtualUser::ReleaseRequest(DWORD dwConnectionId)
{
    CLock lockReq (m_sl);

    CAsyncHTTPRequest* pReq = GetRequestFromConnId(dwConnectionId);
    if(pReq != NULL)
    {
        ReleaseRequest(pReq);
    }
    else
    {
        assert(0);
    }
}

// --------------------------------------------------------------------------------
// ReleaseRequest
//
// Release the connection slot associated with the connection slot pointed at by
// pReq.
// --------------------------------------------------------------------------------
void CAsyncVirtualUser::ReleaseRequest(CAsyncHTTPRequest* pReq)
{
    assert(pReq != NULL);
    CLock lock(m_sl);

    pReq->OnRequestClosed();
}

// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------
DWORD CAsyncVirtualUser::DisconnectClient(DWORD dwConnectionId)
{
    CLock lock(m_sl);

    // Search the slot owned by dwConnectionId.
    CAsyncHTTPRequest* pReq = GetRequestFromConnId(dwConnectionId);
    if(pReq != NULL)
    {
        if(     pReq->GetStatus() != CAsyncHTTPRequest::CLIENT_INITIAL
            &&  pReq->GetStatus() != CAsyncHTTPRequest::CLIENT_CONNECTED)
        {
            // Should not happen (must first terminate the request). Otherwise, chances are status changed inbetween.
            // May result in events from a previous request being considered as results of the next one?
            if(GetTrace() && GetTrace()->IsTrace("errors"))
            {
                GetTraceStream(GetTrace()) << "DisconnectClient: ERROR. Wrong request status: " << pReq->GetStatus() << ".\r\n";
                ReleaseTraceStream();
            }

            assert(0);
        }

        m_dwDisconnectToWait = 1;

        if(pReq->Disconnect() == ERROR_IO_PENDING)
        {
            m_execPhase = EXEC_WAIT_DISCONNECTS;

            return TOF_RES_ASYNC_STARTED;
        }
        else
            return TOF_RES_OK;
    }
    else
    {
        // WARNING: Did not find that connectId. Probably did a Disconnect on a connection that does not exist.
        if(GetTrace() && GetTrace()->IsTrace("connection"))
        {
            GetTraceStream(GetTrace()) << "WARNING: DISCONNECT CLIENT: could not find a connection with id " << dwConnectionId << ". Already disconnected?\r\n";
            ReleaseTraceStream();
        }

        return TOF_RES_OK;
    }
}

// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------
DWORD CAsyncVirtualUser::DisconnectAllClients()
{
    CLock lock(m_sl);

    m_dwDisconnectToWait = 0;

    CAsyncHTTPRequest* pCurrRequest;

    for(DWORD dw=0 ; dw<m_requests.size() ; ++dw)
    {
        pCurrRequest = (CAsyncHTTPRequest*)m_requests[dw];
        if(pCurrRequest->GetConnectionId() != ULONG_MAX)
        {
            if(    pCurrRequest->GetStatus() != CAsyncHTTPRequest::CLIENT_INITIAL
                && pCurrRequest->GetStatus() != CAsyncHTTPRequest::CLIENT_CONNECTED)
            {
                // Should not happen (must first terminate the request). Otherwise, chances are status changed inbetween.
                // May result in events from a previous request being considered as results of the next one?
                if(GetTrace() && GetTrace()->IsTrace("errors"))
                {
                    GetTraceStream(GetTrace()) << "DisconnectAllClients: ERROR. Wrong request status: " << pCurrRequest->GetStatus() << ".\r\n";
                    ReleaseTraceStream();
                }

                assert(0);
            }

            if(GetTrace() && GetTrace()->IsTrace("connection"))
            {
                GetTraceStream(GetTrace()) << "DisconnectAllClients: disconnecting connection " << pCurrRequest->GetConnectionId() << ".\r\n";
                ReleaseTraceStream();
            }

            ++m_dwDisconnectToWait;

            if(pCurrRequest->Disconnect() == ERROR_IO_PENDING)
            {
                m_execPhase = EXEC_WAIT_DISCONNECTS;
            }
        }
    }

    if(m_execPhase == EXEC_WAIT_DISCONNECTS)
        return TOF_RES_ASYNC_STARTED;
    else
        return TOF_RES_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//  Code which purpose is synchronisation.
//
///////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------------
// Return FALSE if connection dwClientConnId is still busy (thus caller must wait), 
// or TRUE if it's free or not created yet. Note CLIENT_CONNECTED is not a busy 
// status.
// --------------------------------------------------------------------------------
BOOL CAsyncVirtualUser::SynchroniseHTTPClient(DWORD dwClientConnId)
{
    assert(dwClientConnId != ULONG_MAX);
    BOOL bIsUnderProcess = FALSE;

    CLock lock(m_sl);

    // Search the slot owned by dwConnectionId.
    CAsyncHTTPRequest* pReq = GetRequestFromConnId(dwClientConnId);
    if(pReq != NULL)
    {
        CAsyncHTTPRequest::EClientStateType status = pReq->GetStatus();

        if(    status != CAsyncHTTPRequest::CLIENT_INITIAL
            && status != CAsyncHTTPRequest::CLIENT_CONNECTED)
        {
            if(GetTrace() && GetTrace()->IsTrace("sync"))
            {
                GetTraceStream(GetTrace()) << "SYNC HTTP CLIENT Initiating a wait operation on connection " << dwClientConnId << "\r\n";
                ReleaseTraceStream();
            }

            m_dwWaitedConnId = dwClientConnId;
            m_execPhase = EXEC_HTTP_SYNC;

            return FALSE;
        }
    }

    // dwClientConnId is sleeping.
    return TRUE;
}

// --------------------------------------------------------------------------------
// Return TRUE if no request is busy anymore.
// --------------------------------------------------------------------------------
BOOL CAsyncVirtualUser::SynchroniseAllHTTPClients()
{
    CLock lock(m_sl);

    if(m_dwPendingRequests == 0)
    {
        if(GetTrace() && GetTrace()->IsTrace("request"))
        {
            GetTraceStream(GetTrace()) << "SYNC ALL HTTP CLIENTS: all requests completed.\r\n";
            ReleaseTraceStream();
        }
        return TRUE;
    }
    else
    {
        if(GetTrace() && GetTrace()->IsTrace("request"))
        {
            GetTraceStream(GetTrace()) << "SYNC ALL HTTP CLIENTS: " << m_dwPendingRequests << " requests under process.\r\n";
            ReleaseTraceStream();
        }
        m_execPhase = EXEC_ALL_HTTP_SYNC;
        return FALSE;
    }
}

// --------------------------------------------------------------------------------
// Ask the timer object to warn us when the specified delay is elapsed.
// --------------------------------------------------------------------------------
BOOL CAsyncVirtualUser::StartEventTimer(DWORD dwNbMilliSeconds)
{
    status = MDB_SLEEP;

    assert(m_execPhase == EXEC_RUNNING || m_execPhase == EXEC_INIT);

    m_execPhase = EXEC_WAIT_TIMER;
    if(GetTrace() && GetTrace()->IsTrace("sync"))
    {
        GetTraceStream(GetTrace()) << "START TIMER Initiating a " << dwNbMilliSeconds*timefac/100 << " milliseconds wait operation for user " << m_dwManagerNumber << ".\r\n";
        ReleaseTraceStream();
    }
    m_pTimerQueue->Push((IocpEventDispatcher*) this, dwNbMilliSeconds*timefac/100);

    return TRUE;
}

void CAsyncVirtualUser::CancelQuery(DWORD dwConnectionId)
{
    // Search the slot owned by dwConnectionId.
    CLock lock(m_sl);
    RequestsType::iterator itEnd = m_requests.end();
    for(RequestsType::iterator it = m_requests.begin(); it != itEnd; ++it)
    {
        CAsyncHTTPRequest *pCurrReq = static_cast<CAsyncHTTPRequest*>(*it);
        if(pCurrReq->GetConnectionId() == dwConnectionId)
        {
            pCurrReq->CancelRequest();

            return;
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
//
//  Code which purpose is script execution.
//
///////////////////////////////////////////////////////////////////////////////

// Inspired from TEThread::StartScript.
void CAsyncVirtualUser::StartScript(const string& fullscriptname)
{
    string scriptname = strip_extension(filename_base(string(fullscriptname)));
    set_file(scriptname);

    m_ScriptNameStack.push(scriptname);
    assert(m_ContextDataStack.top() != NULL);

    m_ContextDataStack.top()->exit_called_flag = FALSE;   // TODO : not in use anymore?

#ifndef CYRANO_TP_STANDALONE
    // Register script for Variable factory.
    string Pfactory_regname = Repository->GetPrimaryTestRunString();
    string factory_regname = Repository->GetTestRunString();

    CORBA::Object_var obj;
    obj = pCServer->FindObjectWithName(TMVarFactory::CLASS_NAME, Pfactory_regname.c_str());
    TMVarFactory_var TMVF = TMVarFactory::_narrow( obj.inout());
    CORBA::String_var cb_scriptname = CORBA::string_dup(ScriptName().c_str());
    TMVF->RegisterScript(cb_scriptname.in());

    if(Pfactory_regname != factory_regname) //if the two TestManager Factories are not the same one
    {
        obj = pCServer->FindObjectWithName(TMVarFactory::CLASS_NAME, factory_regname.c_str());
        TMVF = TMVarFactory::_narrow( obj.inout());
        TMVF->RegisterScript(cb_scriptname.in());
    }
#endif

    if(CreateTask(m_szMainTOFFile) == FALSE)
    {
        // Error. Cannot start VU.
        assert(0);
        return;
    }

    ProcessTof();   // Nb: results handled in ProcessTof().
}



///////////////////////////////////////////////////////////////////////////////
// @mfunc Call another script within this thread
// @parm  ScriptName 
// @parm  parameters to be passed to script
// @rdesc -1 for failure
//
// Modifies the calling environment in the context in order to handle the new script.
// Current environment is stored in order to allow to continue with the calling script
// when this one is over.

// Inspired by TEThread::CallScript.
int CAsyncVirtualUser::CallScript(const string& name)
{
    // Get parameter data before we stack up.
    const list<CECExprDsc>* varExprListP = m_paramsInfo.ExprListP();
    if(varExprListP->empty() == FALSE)
    {
        BOOL eval = m_paramsInfo.EvaluateValues(*this, m_Params);
        if(!eval)
            throw mdbParseException("Unable to Evaluate Parameter values for Call Script");
    }

    // Stack up member data.
    CScriptContextData* pData   = new CScriptContextData(m_dwParallelism);
    CScriptContextData* pTop    = CURR_SCRIPT_CTXT_DATA;
    pData->phase                = pTop->phase;
    pData->phasetxt             = "Start New Script";
    pData->flags                = pTop->flags;
    pData->oi                   = -1;

    // Add threadwide variables from previous varlist to this one.
    pData->varlist.add_vars(pTop->varlist.get_threadwide_vars());
    
    m_ContextDataStack.push(pData);
    m_ScriptNameStack.push(name);

    if(CreateTask(name) == FALSE)
        return TOF_RES_FAILED;

    // get parameter data
    // run task, Note: this calls the StartScript and EndScript methods in 
    // derived TContext classes for specific setup/clear down
    int iRet;   // Return value, -1 indicates an error.

    try
    {
        iRet = ProcessTof();
    }
    catch(const DetachThreadException& d)//if thread detatching pass on up to top
    {
        throw d;
    }
    catch(...)
    {
        // Catch all exceptions for now and treat as pass on to error handling to ensure neater closedown.
        iRet = -1;
    }

    return iRet;
}

BOOL CAsyncVirtualUser::CreateTask(const string &szFileName)
{
    if(GetVerbose(AUDITMSG))
        log("Start Script", AUDITMSG);

    // Get new file based upon szFileName.

    // Bodge to stop unnecessary checking out of files from repository.

    // Adjust file name.
    string szScriptName = strip_extension(filename_base(szFileName));

#ifndef CYRANO_TP_STANDALONE
    string tmpString = g_pTExecuter->GetCurrentDir() + FN_PATHSEP + szScriptName + ".tof";
#else
    string tmpString = g_pTExecuter->GetTofFileDir() + FN_PATHSEP + szScriptName + ".tof";
#endif

    set_file(tmpString);    // Push latest FULL file name onto stack.

    string ScriptFilename;

    // Put a mutex lock around the code to retrieve the file.
    CLock guard(s_pCacheSl);

#if defined TH_POOL_VER && defined CYRANO_TP_STANDALONE
    ScriptFilename = tmpString;
#else
    if(((CAsyncVirtualUserController*)m_pController)->get_cache().doesTaskExist(tmpString) == FALSE)
    {
        if(GetTrace() && GetTrace()->IsTrace("Debug-Generic"))
        {
            GetTraceStream(m_pTrace) << "CAsyncVirtualUser::CallScript: Get the File: " << szFileName << ".\r\n";
            ReleaseTraceStream();
        }

        try
        {
            ScriptFilename = Repository->GetTofFile(szScriptName + ".tof");
        }
        catch(...)
        {
            CORBA::String_var what = (string("Unable to retrieve File: ") + szFileName).c_str();
            CORBA::String_var where = "TestExecuterImpl::ExecuteThread";
#   ifndef CYRANO_TP_STANDALONE
	    	throw CyrProcessException(what, where, -1);
#   else
    		throw CyrProcessException(what.c_str(), where.c_str(), -1);
#   endif
        }
    }
    else
    {
        ScriptFilename = tmpString;
    }
#endif

    if(ScriptFilename.empty() == TRUE)
    {
        // Check in case nothing is returned.
        CORBA::String_var where = "TestExecuterImpl::ExecuteThread";
        CORBA::String_var what  = "GetTofFile Failed";
#ifndef CYRANO_TP_STANDALONE
		throw CyrProcessException(what, where, -1);
#else
		throw CyrProcessException(what.c_str(), where.c_str(), -1);
#endif
    }


    //// Create new task
    try
    {
        // Create new task, replacing current.
        CURR_SCRIPT_CTXT_DATA->m_pTaskScript = ((CAsyncVirtualUserController*)m_pController)->get_cache().makeTask(ScriptFilename);
        assert(CURR_SCRIPT_CTXT_DATA->m_pTaskScript != NULL);
        SetScript(CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetScript());
    }
    catch(...)
    {
        if(GetVerbose(ERRMSG))
            log("Error creating new Script Task for " + ScriptFilename, ERRMSG);

        return FALSE;
    }


    //// Initialise the current script status. Used to be done in TScript::run.
    // fill out the context with this scripts variable definitions
    declarations(CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetScript()->GetTrecDefs());

    // Open any external datafiles
    //pCtxt.open_datafiles();   // TODO?

    // Clean out the OpenObjects stack
    while (!OpenObjects().empty())
    {
        OpenObjects().pop();
    }

    ClearMask(~CTXT_DEFAULT);

    // Get record count for script. This must be done at least once
    // after file is loaded to initialize record holder iterator array.
    SetScriptLength(CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetScript()->length());
    SetScriptPos(0);

    return TRUE;
}


// Called when we are over with the called script.
// Set the environment for the calling script again.
// TODO : look at TEThread::EndScript() + ExitCommand.
void CAsyncVirtualUser::EndScript()
{
    if(GetVerbose(AUDITMSG))
        log("End Script", AUDITMSG);

    if(!m_ContextDataStack.top()->exit_called_flag)
    	ExitCommand(FALSE, "");

    if(m_ContextDataStack.size() == 1)
    {
        // The main script is over. Signal the framework we are over, and release memory.
        m_pController->OnAllRequestsFinished(m_dwManagerNumber);

        CLock lock(m_sl);
        for(RequestsType::iterator it = m_requests.begin(); it != m_requests.end(); ++it)
        {
            if(((CAsyncHTTPRequest*)*it)->GetConnectionId() != ULONG_MAX)
                ReleaseRequest((CAsyncHTTPRequest*)*it);
        }
    }
    else    // m_ContextDataStack.size() > 1
    {
        // Do not delete the first item, for algorithm reasons (Cf. TContext...).
        CScriptContextData* pCtxtData = CURR_SCRIPT_CTXT_DATA;

        // Pop member data.
        delete pCtxtData;

        m_ContextDataStack.pop();
        m_ScriptNameStack.pop();

        // Set m_execPhase to running because that's the calling script's value.
        m_execPhase = EXEC_RUNNING;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Evaluate parameters for Call Script Entry Command
// string of comma separated variable names to be set
// 0 for success
int CAsyncVirtualUser::Entry(const TtValList &varPrmListP)
{
    // Get pointer to parameter descriptor list for formal parameters.
    const list<CECExprDsc>* varExprListP = varPrmListP.ExprListP();

    //////////////////////
    //////////////////////
    if (m_paramsInfo.ExprListP()->empty() == TRUE)
    {
        set_variable(varExprListP->begin()->VarNam(), "DEFAULT");   // Set parameter variable	
        return 0;
    }
    //////////////////////
    //////////////////////

    // Get pointer to parameter descriptor list for actual parameters.
    const list<CECExprDsc>* exprListP = m_paramsInfo.ExprListP();
    list<CECExprDsc>::const_iterator dscItr = exprListP->begin();

    // Iterate through the formal parameter list, evaluating the actual parameter
    // values and setting the corresponding parameter variables.
    string prmVal;  // Evaluated parameter value
    string varIdx;  // Variable index
    list<string>::const_iterator ParamValIter = m_Params.begin();
    for (list<CECExprDsc>::const_iterator varDscItr = varExprListP->begin();
        varDscItr != varExprListP->end();
        ++varDscItr, ++ParamValIter)
    {
        // Get pointer to the descriptor for the actual parameter and
        // report an error if there isn't one.
        if (dscItr == exprListP->end()) 
            throw mdbParseException("Required Parameter Missing");// Return error if no actual parameter

        CECExprDsc prmValDsc = *dscItr;			// Get actual parameter descriptor
        ++dscItr;					// Point to next actual parameter

        // Get pointer to the descriptor for the formal parameter and
        // return failure, if the formal and actual parameters do NOT have the same data type.
        CECExprDsc varPrmDsc = *varDscItr;		// Get variable parameter descriptor
        if (prmValDsc.DTyp() != varPrmDsc.DTyp()) 
            throw mdbParseException("Paramter type mismatch");

        // Put the actual parameter value into the corresponding parameter variable.
        list<CECOperand>::const_iterator varExprItr = varPrmDsc.ExprIt();
        if (!(varPrmListP.EvaluateVarIdx(*this, &varExprItr, &varIdx))) 
            throw mdbParseException("Failed to Evaluate");

        prmVal = *ParamValIter;//cooresponding value in list
        set_variable(varPrmDsc.VarNam(), prmVal, varIdx);   // Set parameter variable	

    } // End of for

    return 0;
}

BOOL CAsyncVirtualUser::TryAcquireSemaphore(const string &s_name, unsigned l_scope)
{
#ifndef CYRANO_TP_STANDALONE

    if(m_szCurrentSemaphoreName != s_name)
    {
        if(m_pCurrentSemaphore != NULL)
            delete m_pCurrentSemaphore;

        m_pCurrentSemaphore = new CECSemaphore (s_name, l_scope);
    }

    BOOL bRes = m_pCurrentSemaphore->tryacquire();
    if(bRes == TRUE)
    {
        // No need to keep trace of that semaphore.
        delete m_pCurrentSemaphore;
        m_pCurrentSemaphore = NULL;
    }

    return bRes;

#else

    return TRUE;

#endif
}

#ifndef CYRANO_TP_STANDALONE

void CAsyncVirtualUser::GetVUInfo(ThreadInfo &tI) const
{
    static char *szDumy = "Starting";

    tI.UIdent       = m_ThreadID.c_str();
    tI.id           = intid;    // TODO: what's the use of this?
    if(m_ScriptNameStack.size() > 0)
        tI.ScriptName = m_ScriptNameStack.top().c_str();
    else
        tI.ScriptName = szDumy;
    tI.Status       = (m_execPhase == EXEC_OVER) ? SUSPENDED : ACTIVE;
    tI.StartTime    = m_dwStartTime;
    tI.NoteText     = m_noteText.c_str();
}

#endif

//// --------------------------------------------------------------------------------
//// TContext virtual methods.
//// --------------------------------------------------------------------------------
// Inspired from TEThread.
void CAsyncVirtualUser::CleanUp()
{
    //TODO: not used?
/*
    set_phase_text("CleanUp");

    if(have_cleaned_up)
        return; // Should only do this once per thread

    // Should be explicit but there is an implicit call in the destrctor for aborted threads
    //decrement number of threads running TBD this should have some global mutex locking
    int no_threads=atoi(get_variable(THREAD_COUNT_VAR,true).c_str());
    no_threads--;
    set_variable(THREAD_COUNT_VAR,int_to_string(no_threads));

    ExitCommand(false,"");
#ifndef CYRANO_TP_STANDALONE
    CTLockGuard<CTMutex> guard(m_ThreadMap_mutex);
    ThreadMap_t::size_type sz=m_tethreadcontrol->ThreadMap.erase(m_ThreadID);
    if(sz!=1)//did not remove 1 and only 1 item
    {
      string Errmsg="ThreadInfo for ThreadID :"+m_ThreadID +" Could not be erased properly";
      throw mdbParseException(Errmsg);
    }
#endif
    set_phase_text(TP_PHASE_NONE);
    have_cleaned_up=true;
    */
}

//Call thread type specific method for setting note Text.
void CAsyncVirtualUser::SetNoteText(const string& msg)
{
    m_noteText = msg;
}

string CAsyncVirtualUser::get_test_name()
{
    return m_szTestName;
}

#ifndef CYRANO_TP_STANDALONE

ThreadDef CAsyncVirtualUser::GetThreadDef()
{
    return m_ThreadDef;
}

#endif

// Clear up this Script on exit from script with/without keep alive options set
int CAsyncVirtualUser::ExitCommand(bool keepalive, const string& retvar)
{
    if(m_ContextDataStack.top()->exit_called_flag == TRUE)
	    return 0;   // Have already exited don't try and do it again

    logr->AddTrace(DBGSTD,DETAILLEVEL, "CAsyncVirtualUser::ExitCommand");

#ifndef CYRANO_TP_STANDALONE

    // Deregister ScriptObject from TestManagerVariableFactories for
    string Pfactory_regname = Repository->GetPrimaryTestRunString();
    string factory_regname  = Repository->GetTestRunString();

    CORBA::Object_var obj = pCServer->FindObjectWithName(TMVarFactory::CLASS_NAME, Pfactory_regname.c_str());
    TMVarFactory_var TMVF = TMVarFactory::_narrow( obj.inout());

    if(!CORBA::is_nil(TMVF))
    {
        CORBA::String_var cb_scriptname = CORBA::string_dup(ScriptName().c_str());
//TODO: crashed!?!        TMVF->DeregisterScript(cb_scriptname.in(), keepalive);

        if(Pfactory_regname != factory_regname) // If the two TestManager Factories are not the same one
        {
            obj = pCServer->FindObjectWithName(TMVarFactory::CLASS_NAME, factory_regname.c_str());
            TMVF = TMVarFactory::_narrow( obj.inout());
            TMVF->DeregisterScript(cb_scriptname.in(), keepalive);
        }

        TContext::ExitCommand(keepalive, retvar);
        m_ContextDataStack.top()->exit_called_flag = true;  // Have explicitly called exit in Script
    }
    else
    {
	    logr->AddTrace(DBGSTD,DETAILLEVEL,"TestManager Variable Factory Unreachable");
    }
    return 0;   //Got here without throwing an exception so must be successful

#else
    m_ContextDataStack.top()->exit_called_flag = true;  // Have explicitly called exit in Script
    return 0;

#endif
}


///////////////////////////////////////////////////////////////////////////////
// Terminates the current test case, as either passed or failed depending
// on the status of the flag passed in.
//
// @rdesc   integer 0 if the test case is terminated successfully, otherwise
//	    false.
// @parm    int | result | Flag determining the state of termination
//	    (TOF_TCTRES_PASS/FAIL).
// @parm    description of TestCase passed or failed (taken from start test case
//	    record at prepare time).
int CAsyncVirtualUser::TCterminate(int result, string desc)
{
	if(result == TOF_TCTRES_FAIL)
	    ++m_dwTestCaseFail;
	else
	    ++m_dwTestCasePass;

	return TContext::TCterminate(result, desc);
}

// Read the statistics verbose level flag. Tell the request structures what they're gonna have to do of the stats.
BOOL CAsyncVirtualUser::MustDumpStats() // TODO: should be const.
{
	string zStatVerbLevel;
    try
    {
        zStatVerbLevel = get_variable("CYR_HTTP-STAT-LV");
    }
    catch(...)
    {
        // The case of exception means variable not defined. Should use the default mode which is the low one.
        zStatVerbLevel = 'L';
    }

    BOOL bWatchStats;
    if(zStatVerbLevel[0] == 'N' || (zStatVerbLevel[0] == 'L' && GetManagerNumber() >= 10))
        bWatchStats = FALSE;
    else
        bWatchStats = TRUE;

    return bWatchStats;
}
