/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// CECDataName.cpp
// ---------------
//
//    This is the implementation file for the CECDataName class
//
//////////////////////////////////////////////////////////////////////////////

//
//    This is the cpp file for the CECDataName Class.
//    Specialise the ***SECMaskEdit class by extending the allowable masks
//    to include the new character 'V' to specify a VTEST name
//    A valid VTEST dataname comprises between 1 and 16 alphanumeric
//    underscore or hyphen characters. The first character must be
//    alphabetic; spaces are not allowed; two adjacent underscores
//    or hyphens are not allowed; and neither is a trailing underscore
//    or hyphen.
//
//      The last part of the requirement (ie trailing underscore or hyphen)
//      will have to be implemented later.
//
//    This interpretation is implemented by over-riding the virtual function
//    IsCharValid.
//
//
#include "stdafx.h"
#include "CECDataName.h"

//
//  PSL Macros.
//
extern "C"
{
#include <psl_macros.h>
}
//
//  Function Prototypes.
//
extern "C"
{
#include <psl_rtndef.h>         // PSL Routines 
}


//
// Default constructor
// ===================
//

CECDataName::CECDataName    ( )
        : m_bSubClassed( FALSE ),
        m_bInsert( FALSE )
{
    m_nDataNameLength = 16;
    m_strSymbols = "-_";
    m_strValidChars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-";

}

//
// Destructor
// ==========
//
CECDataName::~CECDataName()
{}


BEGIN_MESSAGE_MAP(CECDataName, CEdit)
//{{AFX_MSG_MAP(CECDataName)
ON_WM_CHAR()
ON_WM_KEYDOWN()
//}}AFX_MSG_MAP
END_MESSAGE_MAP()

//
// ValidDataName
// =============
//
// Evaluate whether the dataname 'strDataName' is a valid VTEST dataname.
//
// Returns TRUE to indicate a valid dataname
//          FALSE to indicate an invalid dataname
BOOL CECDataName::ValidDataName(const CString&  strDataName)
{
    char                chDataName[20];
    short               nSize;
    CString             strLocal;
    short               bRes;

    //
    // Make a local copy of dataname
    strLocal = strDataName;
    strcpy( chDataName, strLocal.GetBuffer(0));

    //
    // Call routine
    nSize = (short) strlen(chDataName);
    CHK_DATANAME( chDataName, &nSize, &bRes);

    //
    // decode the result
    if ( bRes == 0)
    {
        //
        // valid
        return TRUE;
    }
    else
    {
        // invalid
        return FALSE;
    }
}

/////////////////////////////////////////////////////////////////////////////
// CECDataName message handlers
//

//
// SubclassEdit
// ============
//
// This function subclasses the dialog control nID to CEColourEditBox.
//
// Here we over-ride SubclassEdit to set a data name mask.
//
BOOL CECDataName::SubclassEdit (UINT nID, CWnd* pParent, WORD wParseStyle)
{
    //
    // Update internal state
    m_bSubClassed = TRUE;

    wParseStyle; //unused
    SubclassDlgItem( nID, pParent);
    return TRUE;
}

//
// OnChar
// ======
//
// Keyboard message handler called when a keystroke translates into a nonsystem character
//
// For each character we receieve we test whether by inserting that character
// we still have a valid VTEST dataname.
//
// If the character invalidates the VTEST dataname then a beep is made, and we do not
// modify the contents of the edit box.
void CECDataName::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags)
{
    //
    // Retrieve position of cursor.
    DWORD dwStartPos = GetSel();

    //
    // Which is contained in the low order word returned by GetSel
    WORD wStartPos = LOWORD(dwStartPos);

    TCHAR ch = (char) nChar;

    if ( IsCharValid( wStartPos, ch) )
    {
        //
        // force upper-case
        nChar = toupper(nChar);
        _AFX_THREAD_STATE* pThreadState = AfxGetThreadState();
        pThreadState->m_lastSentMsg.wParam = nChar;

        //
        // Process char
        CEdit::OnChar( nChar, nRepCnt, nFlags);
    }
    else
    {
        MessageBeep( MB_ICONEXCLAMATION );
    }
}

//
// IsCharValid
// ===========
//
// The user has typed the character 'ch' at position 'nPos'
// Is the dataname still valid AND is 'ch' allowable at position 'nPos'
BOOL CECDataName::IsCharValid(int nPos,  TCHAR& ch)
{
    ASSERT_VALID(this);
    BOOL bRet = FALSE;
    TCHAR chSpread = ('a' - 'A');
    BOOL bPrevCharSymbol;       // is previous char a symbol
    BOOL bNextCharSymbol;       // is next char a symbol
    char chNextChar;
    char chPrevChar;

    //
    // If the key has already been processed
    if (m_bKeyProcessed)
        return TRUE;

    CString strCurrentText;
    GetWindowText(strCurrentText);
    //
    // If attempting to enter more than allowed characters, SCREAM
    if( strCurrentText.GetLength() == m_nDataNameLength)
        return FALSE;

    //
    // Validate VTEST data names
    if ( m_strValidChars.Find(ch) != -1 )
    {
        //
        // force upper-case
        if(ch >= 'a' && ch <= 'z')
            ch = (TCHAR) (ch - chSpread);
        //
        // if first character ensure only alphabetical
        if (nPos == 0)
            bRet = (ch >= 'A' && ch <= 'Z');
        else
        {
            if ( m_strSymbols.Find( ch) != -1)
            {
                //
                // If at the last character position of the mask
                // then definitley not OK - current position zero based index
                if( (int)nPos+1 == m_nDataNameLength)
                    return FALSE;
                //
                // if character is symobl, ensure previous/next
                // character is NOT a symbol.
                // if we are in over-write mode the next character
                // is the character at cursor location +1
                // if in insert mode the next character is the
                // character currently under the cursor.
                chPrevChar = strCurrentText .GetAt( nPos-1);
                if (m_bInsert)
                    chNextChar = strCurrentText .GetAt( nPos);
                else
                {
                    if ((nPos+1) < strCurrentText.GetLength())
                        chNextChar = strCurrentText.GetAt( nPos+1);
                    else
                        chNextChar = ' ';
                }
                bPrevCharSymbol = m_strSymbols.Find( chPrevChar) != -1;
                bNextCharSymbol = m_strSymbols.Find( chNextChar) != -1;
                bRet = ((!bPrevCharSymbol) && (!bNextCharSymbol));
            }
            else
                bRet = TRUE;
        } // else clause in nPos==0
    } // if character is allowable

    else
    {
        bRet = FALSE;
    } // else clause of if char allowable.

    return bRet;
}

//
// OnKeyDown
// =========
//
// Message handler invoked whenever a key is pressed.
//
void  CECDataName::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
    m_bKeyProcessed = FALSE;


    if (nChar == VK_INSERT)
    {
        m_bInsert = !m_bInsert;
        m_bKeyProcessed = TRUE;
    }
    else if (nChar == VK_RETURN)
    {
        CString strDataName;
        GetWindowText(strDataName);
        if (CECDataName::ValidDataName(strDataName) == FALSE)
        {
            m_bKeyProcessed = TRUE;
            ::MessageBeep(MB_ICONEXCLAMATION);
        }
    }
    //
    // The following keys are treated as characters, hence they need to be
    // flagged as having been dealt with.
    else if (nChar == VK_BACK)
    {
        m_bKeyProcessed = TRUE;
    }

    if (!m_bKeyProcessed)
    {
        CEdit::OnKeyDown(nChar, nRepCnt, nFlags);
    }
}

//
// SetLength
// =========
//
// Set/Reset the length of the edit field.
void CECDataName::SetLength(const int nLength)
{
    m_nDataNameLength = nLength;
}


//
// GetData
// =======
//
// Function to get the data name.
// TODO - Because we are no longer deriving from SECMaskEdit we can remove
// these two Get/Set data functions and use the UpdateData mechanism.
void CECDataName::GetData(CString& strDataName, BOOL bUnused)
{
    bUnused;

    GetWindowText( strDataName );
}

//
// SetData
// =======
//
// Function to set the data name.
// TODO - Because we are no longer deriving from SECMaskEdit we can remove
// these two Get/Set data functions and use the UpdateData mechanism.
void CECDataName::SetData(CString& strDataName)
{
    SetWindowText( strDataName );
}

//
// PreTranslateMessage
// ===================
//
// Used by class CWinApp to translate window messages before they are dispatched to
// the TranslateMessage and DispatchMessage Windows functions.
//
// If the function returns non-zero it indicates the message has been dealt with and need
// not be dispatched.
//
// We use this mechanism to enforce the requirement that the user can not enter
// a dataname where the last character is a symbol, ('-' or '_')

BOOL CECDataName::PreTranslateMessage(MSG* pMsg)
{
    if ((pMsg->message == WM_KEYDOWN) &&        // message is keypress AND
            (pMsg->wParam == 13))                   // key pressed is a CR
    {
        //
        // Retrieve the dataname
        CString strDataName;
        GetWindowText( strDataName );

        //
        // Check if last char is a symbol
        BOOL     bLastCharSymbol = FALSE;
        if (!strDataName.IsEmpty() )
        {
            CString chLastChar = strDataName.Right( 1 );
            bLastCharSymbol = (m_strSymbols.Find(chLastChar) != -1);
        }

        //
        // Validate dataname not empty AND last char not a symbol.
        if ((strDataName.IsEmpty()) ||
                (bLastCharSymbol))
        {
            MessageBeep( MB_ICONEXCLAMATION );
            return TRUE;
        }
    }

    return CEdit::PreTranslateMessage(pMsg);
}
